from yt.fields.field_info_container import FieldInfoContainer


class StreamFieldInfo(FieldInfoContainer):
    known_other_fields = (
        ("density", ("code_mass/code_length**3", ["density"], None)),
        (
            "dark_matter_density",
            ("code_mass/code_length**3", ["dark_matter_density"], None),
        ),
        ("number_density", ("1/code_length**3", ["number_density"], None)),
        ("pressure", ("dyne/code_length**2", ["pressure"], None)),
        ("specific_thermal_energy", ("erg / g", ["specific_thermal_energy"], None)),
        ("temperature", ("K", ["temperature"], None)),
        ("velocity_x", ("code_length/code_time", ["velocity_x"], None)),
        ("velocity_y", ("code_length/code_time", ["velocity_y"], None)),
        ("velocity_z", ("code_length/code_time", ["velocity_z"], None)),
        ("magnetic_field_x", ("gauss", [], None)),
        ("magnetic_field_y", ("gauss", [], None)),
        ("magnetic_field_z", ("gauss", [], None)),
        (
            "radiation_acceleration_x",
            ("code_length/code_time**2", ["radiation_acceleration_x"], None),
        ),
        (
            "radiation_acceleration_y",
            ("code_length/code_time**2", ["radiation_acceleration_y"], None),
        ),
        (
            "radiation_acceleration_z",
            ("code_length/code_time**2", ["radiation_acceleration_z"], None),
        ),
        ("metallicity", ("Zsun", ["metallicity"], None)),
        # We need to have a bunch of species fields here, too
        ("metal_density", ("code_mass/code_length**3", ["metal_density"], None)),
        ("hi_density", ("code_mass/code_length**3", ["hi_density"], None)),
        ("hii_density", ("code_mass/code_length**3", ["hii_density"], None)),
        ("h2i_density", ("code_mass/code_length**3", ["h2i_density"], None)),
        ("h2ii_density", ("code_mass/code_length**3", ["h2ii_density"], None)),
        ("h2m_density", ("code_mass/code_length**3", ["h2m_density"], None)),
        ("hei_density", ("code_mass/code_length**3", ["hei_density"], None)),
        ("heii_density", ("code_mass/code_length**3", ["heii_density"], None)),
        ("heiii_density", ("code_mass/code_length**3", ["heiii_density"], None)),
        ("hdi_density", ("code_mass/code_length**3", ["hdi_density"], None)),
        ("di_density", ("code_mass/code_length**3", ["di_density"], None)),
        ("dii_density", ("code_mass/code_length**3", ["dii_density"], None)),
    )

    known_particle_fields = (
        ("particle_position", ("code_length", ["particle_position"], None)),
        ("particle_position_x", ("code_length", ["particle_position_x"], None)),
        ("particle_position_y", ("code_length", ["particle_position_y"], None)),
        ("particle_position_z", ("code_length", ["particle_position_z"], None)),
        ("particle_velocity", ("code_length/code_time", ["particle_velocity"], None)),
        (
            "particle_velocity_x",
            ("code_length/code_time", ["particle_velocity_x"], None),
        ),
        (
            "particle_velocity_y",
            ("code_length/code_time", ["particle_velocity_y"], None),
        ),
        (
            "particle_velocity_z",
            ("code_length/code_time", ["particle_velocity_z"], None),
        ),
        ("particle_index", ("", ["particle_index"], None)),
        (
            "particle_gas_density",
            ("code_mass/code_length**3", ["particle_gas_density"], None),
        ),
        ("particle_gas_temperature", ("K", ["particle_gas_temperature"], None)),
        ("particle_mass", ("code_mass", ["particle_mass"], None)),
        ("smoothing_length", ("code_length", ["smoothing_length"], None)),
        ("density", ("code_mass/code_length**3", ["density"], None)),
        ("temperature", ("code_temperature", ["temperature"], None)),
        ("creation_time", ("code_time", ["creation_time"], None)),
        ("age", ("code_time", [], None)),
    )

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        for field in self.ds.stream_handler.field_units:
            if field[0] in self.ds.particle_types:
                continue
            units = self.ds.stream_handler.field_units[field]
            if units != "":
                self.add_output_field(field, sampling_type="cell", units=units)
        setup_magnetic_field_aliases(
            self, "stream", [f"magnetic_field_{ax}" for ax in "xyz"]
        )

    def add_output_field(self, name, sampling_type, **kwargs):
        if name in self.ds.stream_handler.field_units:
            kwargs["units"] = self.ds.stream_handler.field_units[name]
        super().add_output_field(name, sampling_type, **kwargs)



import os
from io import StringIO

import numpy as np

from yt.funcs import mylog


def get_thingking_deps():
    try:
        from thingking.arbitrary_page import PageCacheURL
        from thingking.httpmmap import HTTPArray
    except ImportError:
        raise ImportError(
            "This functionality requires the thingking package to be installed"
        )
    return HTTPArray, PageCacheURL


_types = {
    "int16_t": "int16",
    "uint16_t": "uint16",
    "int": "int32",
    "int32_t": "int32",
    "uint32_t": "uint32",
    "int64_t": "int64",
    "uint64_t": "uint64",
    "float": "float32",
    "double": "float64",
    "unsigned int": "I",
    "unsigned char": "B",
    "char": "B",
}

_rev_types = {}
for v, t in _types.items():
    _rev_types[t] = v
_rev_types["<f8"] = "double"
_rev_types["<f4"] = "float"
_rev_types["<i4"] = "int32_t"
_rev_types["<i8"] = "int64_t"
_rev_types["<u4"] = "uint32_t"
_rev_types["|u1"] = "char"


def _get_type(vtype, tlen=None):
    try:
        t = _types[vtype]
        if tlen is not None:
            t = np.dtype((t, tlen))
        else:
            t = np.dtype(t)
    except KeyError:
        t = eval("np." + vtype)
    return t


def _lstrip(text_list):
    return [t.strip() for t in text_list]


def _get_struct_vars(line):
    spl = _lstrip(line.split(";"))
    multiv = _lstrip(spl[0].split(","))
    ret = _lstrip(multiv[0].split())
    ctype = ret[0]
    vnames = [ret[-1]] + multiv[1:]
    vnames = [v.strip() for v in vnames]
    for vtype in ret[1:-1]:
        ctype += " " + vtype
    num = None
    if len(vnames) == 1:
        if "[" in vnames[0]:
            num = int(vnames[0].split("[")[-1].strip("]"))
            # num = int(re.sub("\D", "", vnames[0]))
    ctype = _get_type(ctype, tlen=num)
    return ctype, vnames


def bbox_filter(left, right, domain_width):
    def myfilter(chunk, mask=None):
        pos = np.array([chunk["x"], chunk["y"], chunk["z"]]).T

        # This hurts, but is useful for periodicity. Probably should check
        # first if it is even needed for a given left/right
        for i in range(3):
            pos[:, i] = np.mod(pos[:, i] - left[i], domain_width[i]) + left[i]

        # Now get all particles that are within the bbox
        if mask is None:
            mask = np.all(pos >= left, axis=1)
            np.logical_and(mask, np.all(pos < right, axis=1), mask)
        else:
            np.logical_and(mask, np.all(pos >= left, axis=1), mask)
            np.logical_and(mask, np.all(pos < right, axis=1), mask)
        return mask

    return myfilter


def sphere_filter(center, radius, domain_width):
    def myfilter(chunk, mask=None):
        pos = np.array([chunk["x"], chunk["y"], chunk["z"]]).T
        left = center - radius

        # This hurts, but is useful for periodicity. Probably should check
        # first if it is even needed for a given left/right
        for i in range(3):
            pos[:, i] = np.mod(pos[:, i] - left[i], domain_width[i]) + left[i]

        # Now get all particles that are within the radius
        if mask is None:
            mask = ((pos - center) ** 2).sum(axis=1) ** 0.5 < radius
        else:
            np.multiply(mask, np.linalg.norm(pos - center, 2) < radius, mask)
        return mask

    return myfilter


def _ensure_xyz_fields(fields):
    for f in "xyz":
        if f not in fields:
            fields.append(f)


def spread_bitsv(ival, level):
    res = np.zeros_like(ival, dtype="int64")
    for i in range(level):
        ares = np.bitwise_and(ival, 1 << i) << (i * 2)
        np.bitwise_or(res, ares, res)
    return res


def get_keyv(iarr, level):
    i1, i2, i3 = (v.astype("int64") for v in iarr)
    i1 = spread_bitsv(i1, level)
    i2 = spread_bitsv(i2, level) << 1
    i3 = spread_bitsv(i3, level) << 2
    np.bitwise_or(i1, i2, i1)
    np.bitwise_or(i1, i3, i1)
    return i1


class DataStruct:
    """docstring for DataStruct"""

    _offset = 0

    def __init__(self, dtypes, num, filename):
        self.filename = filename
        self.dtype = np.dtype(dtypes)
        self.size = num
        self.itemsize = self.dtype.itemsize
        self.data = {}
        self.handle = None

    def set_offset(self, offset):
        self._offset = offset
        if self.size == -1:
            file_size = os.path.getsize(self.filename)
            file_size -= offset
            self.size = float(file_size) / self.itemsize
            assert int(self.size) == self.size

    def build_memmap(self):
        assert self.size != -1
        self.handle = np.memmap(
            self.filename,
            dtype=self.dtype,
            mode="r",
            shape=self.size,
            offset=self._offset,
        )
        for k in self.dtype.names:
            self.data[k] = self.handle[k]

    def __del__(self):
        if self.handle is not None:
            try:
                self.handle.close()
            except AttributeError:
                pass
            del self.handle
            self.handle = None

    def __getitem__(self, key):
        mask = None
        if isinstance(key, (int, np.integer)):
            if key == -1:
                key = slice(-1, None)
            else:
                key = slice(key, key + 1)
        elif isinstance(key, np.ndarray):
            mask = key
            key = slice(None, None)
        if not isinstance(key, slice):
            raise NotImplementedError
        if key.start is None:
            key = slice(0, key.stop)
        if key.stop is None:
            key = slice(key.start, self.shape)
        if key.start < 0:
            key = slice(self.size + key.start, key.stop)
        if key.stop < 0:
            key = slice(key.start, self.size + key.stop)
        arr = self.handle[key.start : key.stop]
        if mask is None:
            return arr
        else:
            return arr[mask]


class RedirectArray:
    """docstring for RedirectArray"""

    def __init__(self, http_array, key):
        self.http_array = http_array
        self.key = key
        self.size = http_array.shape
        self.dtype = http_array.dtype[key]

    def __getitem__(self, sl):
        if isinstance(sl, int):
            return self.http_array[sl][self.key][0]
        return self.http_array[sl][self.key]


class HTTPDataStruct(DataStruct):
    """docstring for HTTPDataStruct"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        HTTPArray, PageCacheURL = get_thingking_deps()
        self.HTTPArray = HTTPArray
        self.pcu = PageCacheURL(self.filename)

    def set_offset(self, offset):
        self._offset = offset
        if self.size == -1:
            # Read small piece:
            file_size = self.pcu.total_size
            file_size -= offset
            self.size = float(file_size) / self.itemsize
            assert int(self.size) == self.size

    def build_memmap(self):
        assert self.size != -1
        mylog.info(
            "Building memmap with offset: %i and size %i", self._offset, self.size
        )
        self.handle = self.HTTPArray(
            self.filename, dtype=self.dtype, shape=self.size, offset=self._offset
        )
        for k in self.dtype.names:
            self.data[k] = RedirectArray(self.handle, k)


class SDFRead(dict):

    _eof = "SDF-EO"
    _data_struct = DataStruct

    def __init__(self, filename=None, header=None):
        r"""Read an SDF file, loading parameters and variables.

        Given an SDF file (see https://bitbucket.org/JohnSalmon/sdf), parse the
        ASCII header and construct numpy memmap array
        access.

        Parameters
        ----------
        filename: string
        The filename associated with the data to be loaded.
        header: string, optional
        If separate from the data file, a file containing the
        header can be specified. Default: None.

        Returns
        -------
        self : SDFRead object
        Dict-like container of parameters and data.


        References
        ----------
        SDF is described here:

            J. K. Salmon and M. S. Warren. Self-Describing File (SDF) Library.
            Zenodo, Jun 2014. URL https://bitbucket.org/JohnSalmon/sdf.

        Examples
        --------

        >>> sdf = SDFRead("data.sdf", header="data.hdr")
        >>> print(sdf.parameters)
        >>> print(sdf["x"])

        """
        self.filename = filename
        if header is None:
            header = filename
        self.header = header
        self.parameters = {}
        self.structs = []
        self.comments = []
        if filename is not None:
            self.parse_header()
            self.set_offsets()
            self.load_memmaps()

    def write(self, filename):
        f = open(filename, "w")
        f.write("# SDF 1.0\n")
        f.write(f"parameter byteorder = {self.parameters['byteorder']};\n")
        for c in self.comments:
            if "\x0c" in c:
                continue
            if "SDF 1.0" in c:
                continue
            f.write(f"{c}")
        for k, v in sorted(self.parameters.items()):
            if k == "byteorder":
                continue
            try:
                t = _rev_types[v.dtype.name]
            except Exception:
                t = type(v).__name__
            if t == str.__name__:
                f.write(f'parameter {k} = "{v}";\n')
            else:
                f.write(f"{t} {k} = {v};\n")

        struct_order = []
        for s in self.structs:
            f.write("struct {\n")
            to_write = []
            for var in s.dtype.descr:
                k, v = var[0], _rev_types[var[1]]
                to_write.append(k)
                f.write(f"\t{v} {k};\n")
            f.write("}[%i];\n" % s.size)
            struct_order.append(to_write)
        f.write("#\x0c\n")
        f.write("# SDF-EOH\n")
        return struct_order, f

    def __repr__(self):
        disp = f"<SDFRead Object> file: {self.filename}\n"
        disp += "parameters: \n"
        for k, v in self.parameters.items():
            disp += f"\t{k}: {v}\n"
        disp += "arrays: \n"
        for k, v in self.items():
            disp += f"\t{k}[{v.size}]\n"
        return disp

    def parse_header(self):
        """docstring for parse_header"""
        # Pre-process
        ascfile = open(self.header)
        while True:
            l = ascfile.readline()
            if self._eof in l:
                break

            self.parse_line(l, ascfile)

        hoff = ascfile.tell()
        ascfile.close()
        if self.header != self.filename:
            hoff = 0
        self.parameters["header_offset"] = hoff

    def parse_line(self, line, ascfile):
        """Parse a line of sdf"""

        if "struct" in line:
            self.parse_struct(line, ascfile)
            return

        if "#" in line:
            self.comments.append(line)
            return

        spl = _lstrip(line.split("="))
        vtype, vname = _lstrip(spl[0].split())
        vname = vname.strip("[]")
        vval = spl[-1].strip(";")
        if vtype == "parameter":
            self.parameters[vname] = vval
            return
        elif vtype == "char":
            vtype = "str"

        try:
            vval = eval("np." + vtype + f"({vval})")
        except AttributeError:
            if vtype not in _types:
                mylog.warning("Skipping parameter %s", vname)
                return
            vval = eval("np." + _types[vtype] + f"({vval})")

        self.parameters[vname] = vval

    def parse_struct(self, line, ascfile):
        assert "struct" in line

        str_types = []
        l = ascfile.readline()
        while "}" not in l:
            vtype, vnames = _get_struct_vars(l)
            for v in vnames:
                str_types.append((v, vtype))
            l = ascfile.readline()
        spec_chars = r"{}[]\;\n\\"
        num = l.strip(spec_chars)
        if len(num) == 0:
            # We need to compute the number of records.  The DataStruct will
            # handle this.
            num = "-1"
        num = int(num)
        struct = self._data_struct(str_types, num, self.filename)
        self.structs.append(struct)
        return

    def set_offsets(self):
        running_off = self.parameters["header_offset"]
        for struct in self.structs:
            struct.set_offset(running_off)
            running_off += struct.size * struct.itemsize
        return

    def load_memmaps(self):
        for struct in self.structs:
            struct.build_memmap()
            self.update(struct.data)


class HTTPSDFRead(SDFRead):

    r"""Read an SDF file hosted on the internet.

    Given an SDF file (see https://bitbucket.org/JohnSalmon/sdf), parse the
    ASCII header and construct numpy memmap array
    access.

    Parameters
    ----------
    filename : string
        The filename associated with the data to be loaded.
    header : string, optional
        If separate from the data file, a file containing the
        header can be specified. Default: None.

    Returns
    -------
    self : SDFRead object
        Dict-like container of parameters and data.

    References
    ----------
    SDF is described here:

        J. K. Salmon and M. S. Warren. Self-Describing File (SDF) Library.
        Zenodo, Jun 2014. URL https://bitbucket.org/JohnSalmon/sdf.

    Examples
    --------

    >>> sdf = SDFRead("data.sdf", header="data.hdr")
    >>> print(sdf.parameters)
    >>> print(sdf["x"])

    """

    _data_struct = HTTPDataStruct

    def __init__(self, *args, **kwargs):
        HTTPArray, _ = get_thingking_deps()
        self.HTTPArray = HTTPArray
        super().__init__(*args, **kwargs)

    def parse_header(self):
        """docstring for parse_header"""
        # Pre-process
        ascfile = self.HTTPArray(self.header)
        max_header_size = 1024 * 1024
        lines = StringIO(ascfile[:max_header_size].data[:])
        while True:
            l = lines.readline()
            if self._eof in l:
                break

            self.parse_line(l, lines)

        hoff = lines.tell()
        if self.header != self.filename:
            hoff = 0
        self.parameters["header_offset"] = hoff


def load_sdf(filename, header=None):
    r"""Load an SDF file.

    Given an SDF file (see https://bitbucket.org/JohnSalmon/sdf), parse the
    ASCII header and construct numpy memmap array access. The file can
    be either local (on a hard drive, for example), or remote (on the World
    Wide Web).

    Parameters
    ----------
    filename: string
        The filename or WWW address associated with the data to be loaded.
    header: string, optional
        If separate from the data file, a file containing the
        header can be specified. Default: None.

    Returns
    -------
    sdf : SDFRead object
        Dict-like container of parameters and data.

    References
    ----------
    SDF is described here:

        J. K. Salmon and M. S. Warren. Self-Describing File (SDF) Library.
        Zenodo, Jun 2014. URL https://bitbucket.org/JohnSalmon/sdf.

    Examples
    --------

    >>> sdf = SDFRead("data.sdf", header="data.hdr")
    >>> print(sdf.parameters)
    >>> print(sdf["x"])

    """
    if "http" in filename:
        sdf = HTTPSDFRead(filename, header=header)
    else:
        sdf = SDFRead(filename, header=header)
    return sdf


def _shift_periodic(pos, left, right, domain_width):
    """
    Periodically shift positions that are right of left+domain_width to
    the left, and those left of right-domain_width to the right.
    """
    for i in range(3):
        mask = pos[:, i] >= left[i] + domain_width[i]
        pos[mask, i] -= domain_width[i]
        mask = pos[:, i] < right[i] - domain_width[i]
        pos[mask, i] += domain_width[i]
    return


class SDFIndex:

    """docstring for SDFIndex

    This provides an index mechanism into the full SDF Dataset.

    Most useful class methods:
        get_cell_data(level, cell_iarr, fields)
        iter_bbox_data(left, right, fields)
        iter_bbox_data(left, right, fields)

    """

    def __init__(self, sdfdata, indexdata, level=None):
        super().__init__()
        self.sdfdata = sdfdata
        self.indexdata = indexdata
        if level is None:
            level = self.indexdata.parameters.get("level", None)
        self.level = level

        self.rmin = None
        self.rmax = None
        self.domain_width = None
        self.domain_buffer = 0
        self.domain_dims = 0
        self.domain_active_dims = 0
        self.wandering_particles = False
        self.valid_indexdata = True
        self.masks = {
            "p": int("011" * level, 2),
            "t": int("101" * level, 2),
            "r": int("110" * level, 2),
            "z": int("011" * level, 2),
            "y": int("101" * level, 2),
            "x": int("110" * level, 2),
            2: int("011" * level, 2),
            1: int("101" * level, 2),
            0: int("110" * level, 2),
        }
        self.dim_slices = {
            "p": slice(0, None, 3),
            "t": slice(1, None, 3),
            "r": slice(2, None, 3),
            "z": slice(0, None, 3),
            "y": slice(1, None, 3),
            "x": slice(2, None, 3),
            2: slice(0, None, 3),
            1: slice(1, None, 3),
            0: slice(2, None, 3),
        }
        self.set_bounds()
        self._midx_version = self.indexdata.parameters.get("midx_version", 0)
        if self._midx_version >= 1.0:
            max_key = self.get_key(np.array([2 ** self.level - 1] * 3, dtype="int64"))
        else:
            max_key = self.indexdata["index"][-1]
        self._max_key = max_key

    def _fix_rexact(self, rmin, rmax):

        center = 0.5 * (rmax + rmin)
        mysize = rmax - rmin
        mysize *= 1.0 + 4.0 * np.finfo(np.float32).eps
        self.rmin = center - 0.5 * mysize
        self.rmax = center + 0.5 * mysize

    def set_bounds(self):
        if (
            "x_min" in self.sdfdata.parameters and "x_max" in self.sdfdata.parameters
        ) or (
            "theta_min" in self.sdfdata.parameters
            and "theta_max" in self.sdfdata.parameters
        ):
            if "x_min" in self.sdfdata.parameters:
                rmin = np.array(
                    [
                        self.sdfdata.parameters["x_min"],
                        self.sdfdata.parameters["y_min"],
                        self.sdfdata.parameters["z_min"],
                    ]
                )
                rmax = np.array(
                    [
                        self.sdfdata.parameters["x_max"],
                        self.sdfdata.parameters["y_max"],
                        self.sdfdata.parameters["z_max"],
                    ]
                )
            elif "theta_min" in self.sdfdata.parameters:
                rmin = np.array(
                    [
                        self.sdfdata.parameters["r_min"],
                        self.sdfdata.parameters["theta_min"],
                        self.sdfdata.parameters["phi_min"],
                    ]
                )
                rmax = np.array(
                    [
                        self.sdfdata.parameters["r_max"],
                        self.sdfdata.parameters["theta_max"],
                        self.sdfdata.parameters["phi_max"],
                    ]
                )
            self._fix_rexact(rmin, rmax)
            self.true_domain_left = self.rmin.copy()
            self.true_domain_right = self.rmax.copy()
            self.true_domain_width = self.rmax - self.rmin
            self.domain_width = self.rmax - self.rmin
            self.domain_dims = 1 << self.level
            self.domain_buffer = 0
            self.domain_active_dims = self.domain_dims
        else:
            mylog.debug("Setting up older data")
            rx = self.sdfdata.parameters.get("Rx")
            ry = self.sdfdata.parameters.get("Ry")
            rz = self.sdfdata.parameters.get("Rz")
            a = self.sdfdata.parameters.get("a", 1.0)
            rmin = -a * np.array([rx, ry, rz])
            rmax = a * np.array([rx, ry, rz])
            self.true_domain_left = rmin.copy()
            self.true_domain_right = rmax.copy()
            self.true_domain_width = rmax - rmin

            expand_root = 0.0
            morton_xyz = self.sdfdata.parameters.get("morton_xyz", False)
            if not morton_xyz:
                mylog.debug("Accounting for wandering particles")
                self.wandering_particles = True
                ic_Nmesh = self.sdfdata.parameters.get("ic_Nmesh", 0)
                # Expand root for non power-of-2
                if ic_Nmesh != 0:
                    f2 = 1 << int(np.log2(ic_Nmesh - 1) + 1)
                    if f2 != ic_Nmesh:
                        expand_root = 1.0 * f2 / ic_Nmesh - 1.0
                        mylog.debug("Expanding: %s, %s, %s", f2, ic_Nmesh, expand_root)
                        rmin *= 1.0 + expand_root
                        rmax *= 1.0 + expand_root

            self._fix_rexact(rmin, rmax)
            self.domain_width = self.rmax - self.rmin
            self.domain_dims = 1 << self.level
            self.domain_buffer = (
                self.domain_dims - int(self.domain_dims / (1.0 + expand_root))
            ) / 2
            self.domain_active_dims = self.domain_dims - 2 * self.domain_buffer

        mylog.debug("MIDX rmin: %s, rmax: %s", self.rmin, self.rmax)
        mylog.debug(
            "MIDX: domain_width: %s, domain_dims: %s, domain_active_dims: %s ",
            self.domain_width,
            self.domain_dims,
            self.domain_active_dims,
        )

    def spread_bits(self, ival, level=None):
        if level is None:
            level = self.level
        res = 0
        for i in range(level):
            res |= ((ival >> i) & 1) << (i * 3)
        return res

    def get_key(self, iarr, level=None):
        if level is None:
            level = self.level
        i1, i2, i3 = (v.astype("int64") for v in iarr)
        return (
            self.spread_bits(i1, level)
            | self.spread_bits(i2, level) << 1
            | self.spread_bits(i3, level) << 2
        )

    def spread_bitsv(self, ival, level=None):
        if level is None:
            level = self.level
        return spread_bitsv(ival, level)

    def get_keyv(self, iarr, level=None):
        if level is None:
            level = self.level
        return get_keyv(iarr, level)

    def get_key_slow(self, iarr, level=None):
        if level is None:
            level = self.level
        i1, i2, i3 = iarr
        rep1 = np.binary_repr(i1, width=self.level)
        rep2 = np.binary_repr(i2, width=self.level)
        rep3 = np.binary_repr(i3, width=self.level)
        inter = np.zeros(self.level * 3, dtype="c")
        inter[self.dim_slices[0]] = rep1
        inter[self.dim_slices[1]] = rep2
        inter[self.dim_slices[2]] = rep3
        return int(inter.tobytes(), 2)

    def get_key_ijk(self, i1, i2, i3, level=None):
        return self.get_key(np.array([i1, i2, i3]), level=level)

    def get_slice_key(self, ind, dim="r"):
        slb = np.binary_repr(ind, width=self.level)
        expanded = np.array([0] * self.level * 3, dtype="c")
        expanded[self.dim_slices[dim]] = slb
        return int(expanded.tobytes(), 2)

    def get_ind_from_key(self, key, dim="r"):
        ind = [0, 0, 0]
        br = np.binary_repr(key, width=self.level * 3)
        for dim in range(3):
            ind[dim] = int(br[self.dim_slices[dim]], 2)
        return ind

    def get_slice_chunks(self, slice_dim, slice_index):
        sl_key = self.get_slice_key(slice_index, dim=slice_dim)
        mask = (self.indexdata["index"] & ~self.masks[slice_dim]) == sl_key
        offsets = self.indexdata["base"][mask]
        lengths = self.indexdata["len"][mask]
        return mask, offsets, lengths

    def get_ibbox_slow(self, ileft, iright):
        """
        Given left and right indicies, return a mask and
        set of offsets+lengths into the sdf data.
        """
        mask = np.zeros(self.indexdata["index"].shape, dtype="bool")
        ileft = np.array(ileft, dtype="int64")
        iright = np.array(iright, dtype="int64")
        for i in range(3):
            left_key = self.get_slice_key(ileft[i], dim=i)
            right_key = self.get_slice_key(iright[i], dim=i)
            dim_inds = self.indexdata["index"] & ~self.masks[i]
            mask *= (dim_inds >= left_key) * (dim_inds <= right_key)
            del dim_inds

        offsets = self.indexdata["base"][mask]
        lengths = self.indexdata["len"][mask]
        return mask, offsets, lengths

    def get_ibbox(self, ileft, iright):
        """
        Given left and right indicies, return a mask and
        set of offsets+lengths into the sdf data.
        """
        # print('Getting data from ileft to iright:',  ileft, iright)

        ix, iy, iz = (iright - ileft) * 1j
        mylog.debug("MIDX IBBOX: %s %s %s %s %s", ileft, iright, ix, iy, iz)

        # plus 1 that is sliced, plus a bit since mgrid is not inclusive
        Z, Y, X = np.mgrid[
            ileft[2] : iright[2] + 1.01,
            ileft[1] : iright[1] + 1.01,
            ileft[0] : iright[0] + 1.01,
        ]

        mask = slice(0, -1, None)
        X = X[mask, mask, mask].astype("int64").ravel()
        Y = Y[mask, mask, mask].astype("int64").ravel()
        Z = Z[mask, mask, mask].astype("int64").ravel()

        if self.wandering_particles:
            # Need to get padded bbox around the border to catch
            # wandering particles.
            dmask = X < self.domain_buffer
            dmask += Y < self.domain_buffer
            dmask += Z < self.domain_buffer
            dmask += X >= self.domain_dims
            dmask += Y >= self.domain_dims
            dmask += Z >= self.domain_dims
            dinds = self.get_keyv([X[dmask], Y[dmask], Z[dmask]])
            dinds = dinds[dinds < self._max_key]
            dinds = dinds[self.indexdata["len"][dinds] > 0]
            # print('Getting boundary layers for wanderers, cells: %i' % dinds.size)

        # Correct For periodicity
        X[X < self.domain_buffer] += self.domain_active_dims
        Y[Y < self.domain_buffer] += self.domain_active_dims
        Z[Z < self.domain_buffer] += self.domain_active_dims
        X[X >= self.domain_buffer + self.domain_active_dims] -= self.domain_active_dims
        Y[Y >= self.domain_buffer + self.domain_active_dims] -= self.domain_active_dims
        Z[Z >= self.domain_buffer + self.domain_active_dims] -= self.domain_active_dims

        # print('periodic:',  X.min(), X.max(), Y.min(), Y.max(), Z.min(), Z.max())

        indices = self.get_keyv([X, Y, Z])
        # Only mask out if we are actually getting data rather than getting indices into
        # a space.
        if self.valid_indexdata:
            indices = indices[indices < self._max_key]
            # indices = indices[self.indexdata['len'][indices] > 0]
            # Faster for sparse lookups. Need better heuristic.
            new_indices = []
            for ind in indices:
                if self.indexdata["len"][ind] > 0:
                    new_indices.append(ind)
            indices = np.array(indices, dtype="int64")

        # indices = np.array([self.get_key_ijk(x, y, z) for x, y, z in zip(X, Y, Z)])
        # Here we sort the indices to batch consecutive reads together.
        if self.wandering_particles:
            indices = np.sort(np.append(indices, dinds))
        else:
            indices = np.sort(indices)
        return indices

    def get_bbox(self, left, right):
        """
        Given left and right indicies, return a mask and
        set of offsets+lengths into the sdf data.
        """
        ileft = np.floor((left - self.rmin) / self.domain_width * self.domain_dims)
        iright = np.floor((right - self.rmin) / self.domain_width * self.domain_dims)
        if np.any(iright - ileft) > self.domain_dims:
            mylog.warning(
                "Attempting to get data from bounding box larger than the domain. "
                "You may want to check your units."
            )
        # iright[iright <= ileft+1] += 1

        return self.get_ibbox(ileft, iright)

    def get_nparticles_bbox(self, left, right):
        """
        Given left and right edges, return total
        number of particles present.
        """
        ileft = np.floor((left - self.rmin) / self.domain_width * self.domain_dims)
        iright = np.floor((right - self.rmin) / self.domain_width * self.domain_dims)
        indices = self.get_ibbox(ileft, iright)
        npart = 0
        for ind in indices:
            npart += self.indexdata["len"][ind]
        return npart

    def get_data(self, chunk, fields):
        data = {}
        for field in fields:
            data[field] = self.sdfdata[field][chunk]
        return data

    def get_next_nonzero_chunk(self, key, stop=None):
        # These next two while loops are to squeeze the keys if they are empty.
        # Would be better to go through and set base equal to the last non-zero base.
        if stop is None:
            stop = self._max_key
        while key < stop:
            if self.indexdata["len"][key] == 0:
                # print('Squeezing keys, incrementing')
                key += 1
            else:
                break
        return key

    def get_previous_nonzero_chunk(self, key, stop=None):
        # These next two while loops are to squeeze the keys if they are empty.
        # Would be better to go through and set base equal to the last non-zero base.
        if stop is None:
            stop = self.indexdata["index"][0]
        while key > stop:
            if self.indexdata["len"][key] == 0:
                # print('Squeezing keys, decrementing')
                key -= 1
            else:
                break
        return key

    def iter_data(self, inds, fields):
        num_inds = len(inds)
        num_reads = 0
        mylog.debug("MIDX Reading %i chunks", num_inds)
        i = 0
        while i < num_inds:
            ind = inds[i]
            base = self.indexdata["base"][ind]
            length = self.indexdata["len"][ind]
            # Concatenate aligned reads
            nexti = i + 1
            combined = 0
            while nexti < num_inds:
                nextind = inds[nexti]
                # print(
                #    "b: %i l: %i end: %i  next: %i"
                #    % (base, length, base + length, self.indexdata["base"][nextind])
                # )
                if combined < 1024 and base + length == self.indexdata["base"][nextind]:
                    length += self.indexdata["len"][nextind]
                    i += 1
                    nexti += 1
                    combined += 1
                else:
                    break

            chunk = slice(base, base + length)
            mylog.debug(
                "Reading chunk %i of length %i after catting %i starting at %i",
                i,
                length,
                combined,
                ind,
            )
            num_reads += 1
            if length > 0:
                data = self.get_data(chunk, fields)
                yield data
                del data
            i += 1
        mylog.debug("Read %i chunks, batched into %i reads", num_inds, num_reads)

    def filter_particles(self, myiter, myfilter):
        for data in myiter:
            mask = myfilter(data)

            if mask.sum() == 0:
                continue
            filtered = {}
            for f in data.keys():
                filtered[f] = data[f][mask]

            yield filtered

    def filter_bbox(self, left, right, myiter):
        """
        Filter data by masking out data outside of a bbox defined
        by left/right. Account for periodicity of data, allowing left/right
        to be outside of the domain.
        """

        for data in myiter:
            # mask = np.zeros_like(data, dtype='bool')
            pos = np.array([data["x"].copy(), data["y"].copy(), data["z"].copy()]).T

            DW = self.true_domain_width
            # This hurts, but is useful for periodicity. Probably should check first
            # if it is even needed for a given left/right
            _shift_periodic(pos, left, right, DW)

            # Now get all particles that are within the bbox
            mask = np.all(pos >= left, axis=1) * np.all(pos < right, axis=1)
            # print('Mask shape, sum:', mask.shape, mask.sum())

            mylog.debug(
                "Filtering particles, returning %i out of %i", mask.sum(), mask.shape[0]
            )

            if not np.any(mask):
                continue

            filtered = {ax: pos[:, i][mask] for i, ax in enumerate("xyz")}
            for f in data.keys():
                if f in "xyz":
                    continue
                filtered[f] = data[f][mask]

            # for i, ax in enumerate('xyz'):
            #    #print(left, right)
            #    assert np.all(filtered[ax] >= left[i])
            #    assert np.all(filtered[ax] < right[i])

            yield filtered

    def filter_sphere(self, center, radius, myiter):
        """
        Filter data by masking out data outside of a sphere defined
        by a center and radius. Account for periodicity of data, allowing
        left/right to be outside of the domain.
        """

        # Get left/right for periodicity considerations
        left = center - radius
        right = center + radius
        for data in myiter:
            pos = np.array([data["x"].copy(), data["y"].copy(), data["z"].copy()]).T

            DW = self.true_domain_width
            _shift_periodic(pos, left, right, DW)

            # Now get all particles that are within the sphere
            mask = ((pos - center) ** 2).sum(axis=1) ** 0.5 < radius

            mylog.debug(
                "Filtering particles, returning %i out of %i", mask.sum(), mask.shape[0]
            )

            if not np.any(mask):
                continue

            filtered = {ax: pos[:, i][mask] for i, ax in enumerate("xyz")}
            for f in data.keys():
                if f in "xyz":
                    continue
                filtered[f] = data[f][mask]

            yield filtered

    def iter_filtered_bbox_fields(self, left, right, data, pos_fields, fields):
        """
        This function should be destroyed, as it will only work with units.
        """

        kpcuq = left.in_units("kpccm").uq
        mpcuq = left.in_units("Mpccm/h").uq
        DW = (self.true_domain_width * kpcuq).in_units("Mpc/h")
        if pos_fields is None:
            pos_fields = "x", "y", "z"
        xf, yf, zf = pos_fields
        mylog.debug("Using position fields: %s", pos_fields)

        # I'm sorry.
        pos = (
            mpcuq
            * np.array(
                [
                    data[xf].in_units("Mpccm/h"),
                    data[yf].in_units("Mpccm/h"),
                    data[zf].in_units("Mpccm/h"),
                ]
            ).T
        )

        # This hurts, but is useful for periodicity. Probably should check first
        # if it is even needed for a given left/right
        _shift_periodic(pos, left, right, DW)

        mylog.debug(
            "Periodic filtering, %s %s %s %s",
            left,
            right,
            pos.min(axis=0),
            pos.max(axis=0),
        )
        # Now get all particles that are within the bbox
        mask = np.all(pos >= left, axis=1) * np.all(pos < right, axis=1)

        mylog.debug(
            "Filtering particles, returning %i out of %i", mask.sum(), mask.shape[0]
        )

        if np.any(mask):
            for i, f in enumerate(pos_fields):
                yield f, pos[:, i][mask]

            for f in fields:
                if f in pos_fields:
                    continue
                # print('yielding nonpos field', f)
                yield f, data[f][mask]

    def iter_bbox_data(self, left, right, fields):
        """
        Iterate over all data within a bounding box defined by a left
        and a right.
        """
        _ensure_xyz_fields(fields)
        mylog.debug("MIDX Loading region from %s to %s", left, right)
        inds = self.get_bbox(left, right)
        # Need to put left/right in float32 to avoid fp roundoff errors
        # in the bbox later.
        # left = left.astype('float32')
        # right = right.astype('float32')

        # my_filter = bbox_filter(left, right, self.true_domain_width)
        yield from self.filter_bbox(left, right, self.iter_data(inds, fields))
        # for dd in self.filter_particles(
        #    self.iter_data(inds, fields),
        #    my_filter):
        #    yield dd

    def iter_sphere_data(self, center, radius, fields):
        """
        Iterate over all data within some sphere defined by a center and
        a radius.
        """
        _ensure_xyz_fields(fields)
        mylog.debug("MIDX Loading spherical region %s to %s", center, radius)
        inds = self.get_bbox(center - radius, center + radius)

        yield from self.filter_sphere(center, radius, self.iter_data(inds, fields))

    def iter_ibbox_data(self, left, right, fields):
        mylog.debug("MIDX Loading region from %s to %s", left, right)
        inds = self.get_ibbox(left, right)
        return self.iter_data(inds, fields)

    def get_contiguous_chunk(self, left_key, right_key, fields):

        lbase = 0
        if left_key > self._max_key:
            raise RuntimeError(
                "Left key is too large. Key: %i Max Key: %i" % (left_key, self._max_key)
            )
        right_key = min(right_key, self._max_key)

        left_key = self.get_next_nonzero_chunk(left_key, right_key - 1)
        right_key = self.get_previous_nonzero_chunk(right_key, left_key)

        lbase = self.indexdata["base"][left_key]

        rbase = self.indexdata["base"][right_key]
        rlen = self.indexdata["len"][right_key]

        length = rbase + rlen - lbase
        if length > 0:
            mylog.debug(
                "Getting contiguous chunk of size %i starting at %i", length, lbase
            )
        return self.get_data(slice(lbase, lbase + length), fields)

    def get_key_data(self, key, fields):
        if key > self._max_key:
            raise RuntimeError(
                "Left key is too large. Key: %i Max Key: %i" % (key, self._max_key)
            )
        base = self.indexdata["base"][key]
        length = self.indexdata["len"][key] - base
        if length > 0:
            mylog.debug(
                "Getting contiguous chunk of size %i starting at %i", length, base
            )
        return self.get_data(slice(base, base + length), fields)

    def iter_slice_data(self, slice_dim, slice_index, fields):
        mask, offsets, lengths = self.get_slice_chunks(slice_dim, slice_index)
        for off, l in zip(offsets, lengths):
            data = {}
            chunk = slice(off, off + l)
            for field in fields:
                data[field] = self.sdfdata[field][chunk]
            yield data
            del data

    def get_key_bounds(self, level, cell_iarr):
        """
        Get index keys for index file supplied.

        level: int
            Requested level
        cell_iarr: array-like, length 3
            Requested cell from given level.

        Returns:
            lmax_lk, lmax_rk
        """
        shift = self.level - level
        level_buff = 0
        level_lk = self.get_key(cell_iarr + level_buff)
        level_rk = self.get_key(cell_iarr + level_buff) + 1
        lmax_lk = level_lk << shift * 3
        lmax_rk = ((level_rk) << shift * 3) - 1
        # print(
        #    "Level ",
        #    level,
        #    np.binary_repr(level_lk, width=self.level * 3),
        #    np.binary_repr(level_rk, width=self.level * 3),
        # )
        # print(
        #    "Level ",
        #    self.level,
        #    np.binary_repr(lmax_lk, width=self.level * 3),
        #    np.binary_repr(lmax_rk, width=self.level * 3),
        # )
        return lmax_lk, lmax_rk

    def find_max_cell(self):
        max_cell = np.argmax(self.indexdata["len"][:])
        return max_cell

    def find_max_cell_center(self):
        max_cell = self.find_max_cell()
        cell_ijk = np.array(
            self.get_ind_from_key(self.indexdata["index"][max_cell]), dtype="int64"
        )
        position = (cell_ijk + 0.5) * (self.domain_width / self.domain_dims) + self.rmin
        return position

    def get_cell_data(self, level, cell_iarr, fields):
        """
        Get data from requested cell

        This uses the raw cell index, and doesn't account for periodicity or
        an expanded domain (non-power of 2).

        level: int
            Requested level
        cell_iarr: array-like, length 3
            Requested cell from given level.         fields: list
            Requested fields

        Returns:
            cell_data: dict
                Dictionary of field_name, field_data
        """
        cell_iarr = np.array(cell_iarr, dtype="int64")
        lk, rk = self.get_key_bounds(level, cell_iarr)
        mylog.debug("Reading contiguous chunk from %i to %i", lk, rk)
        return self.get_contiguous_chunk(lk, rk, fields)

    def get_cell_bbox(self, level, cell_iarr):
        """Get floating point bounding box for a given midx cell

        Returns:
            bbox: array-like of shape (3,2)

        """
        cell_iarr = np.array(cell_iarr, dtype="int64")
        cell_width = self.get_cell_width(level)
        le = self.rmin + cell_iarr * cell_width
        re = le + cell_width
        bbox = np.array([le, re]).T
        assert bbox.shape == (3, 2)
        return bbox

    def iter_padded_bbox_data(self, level, cell_iarr, pad, fields):
        """
        Yields data chunks for a cell on the given level
        plus a padding around the cell, for a list of fields.

        Yields:
            dd: A dictionaries of data.

        Example:

        for chunk in midx.iter_padded_bbox_data(
            6, np.array([128]*3), 8.0, ['x','y','z','ident']):

            print(chunk['x'].max())

        """

        _ensure_xyz_fields(fields)
        bbox = self.get_cell_bbox(level, cell_iarr)
        filter_left = bbox[:, 0] - pad
        filter_right = bbox[:, 1] + pad

        # Center cell
        for dd in self.filter_bbox(
            filter_left, filter_right, [self.get_cell_data(level, cell_iarr, fields)]
        ):
            yield dd
            del dd

        # Bottom & Top
        pbox = bbox.copy()
        pbox[0, 0] -= pad[0]
        pbox[0, 1] += pad[0]
        pbox[1, 0] -= pad[1]
        pbox[1, 1] += pad[1]
        pbox[2, 0] -= pad[2]
        pbox[2, 1] = bbox[2, 0]
        for dd in self.filter_bbox(
            filter_left,
            filter_right,
            self.iter_bbox_data(pbox[:, 0], pbox[:, 1], fields),
        ):
            yield dd
            del dd

        pbox[2, 0] = bbox[2, 1]
        pbox[2, 1] = pbox[2, 0] + pad[2]
        for dd in self.filter_bbox(
            filter_left,
            filter_right,
            self.iter_bbox_data(pbox[:, 0], pbox[:, 1], fields),
        ):
            yield dd
            del dd

        # Front & Back
        pbox = bbox.copy()
        pbox[0, 0] -= pad[0]
        pbox[0, 1] += pad[0]
        pbox[1, 0] -= pad[1]
        pbox[1, 1] = bbox[1, 0]
        for dd in self.filter_bbox(
            filter_left,
            filter_right,
            self.iter_bbox_data(pbox[:, 0], pbox[:, 1], fields),
        ):
            yield dd
            del dd

        pbox[1, 0] = bbox[1, 1]
        pbox[1, 1] = pbox[1, 0] + pad[1]
        for dd in self.filter_bbox(
            filter_left,
            filter_right,
            self.iter_bbox_data(pbox[:, 0], pbox[:, 1], fields),
        ):
            yield dd
            del dd

        # Left & Right
        pbox = bbox.copy()
        pbox[0, 0] -= pad[0]
        pbox[0, 1] = bbox[0, 0]
        for dd in self.filter_bbox(
            filter_left,
            filter_right,
            self.iter_bbox_data(pbox[:, 0], pbox[:, 1], fields),
        ):
            yield dd
            del dd

        pbox[0, 0] = bbox[0, 1]
        pbox[0, 1] = pbox[0, 0] + pad[0]
        for dd in self.filter_bbox(
            filter_left,
            filter_right,
            self.iter_bbox_data(pbox[:, 0], pbox[:, 1], fields),
        ):
            yield dd
            del dd

    def get_padded_bbox_data(self, level, cell_iarr, pad, fields):
        """
        Return list of data chunks for a cell on the given level
        plus a padding around the cell, for a list of fields.

        Returns
        -------
            data: list
                A list of dictionaries of data.

        Examples
        --------
        >>> chunks = midx.get_padded_bbox_data(
        ...     6, np.array([128] * 3), 8.0, ["x", "y", "z", "ident"]
        ... )

        """
        _ensure_xyz_fields(fields)

        data = []
        for dd in self.iter_padded_bbox_data(level, cell_iarr, pad, fields):
            data.append(dd)
        return data

    def get_cell_width(self, level):
        return self.domain_width / 2 ** level

    def iter_padded_bbox_keys(self, level, cell_iarr, pad):
        """

        Returns:
            bbox: array-like of shape (3,2)

        """
        bbox = self.get_cell_bbox(level, cell_iarr)

        # Need to get all of these
        low_key, high_key = self.get_key_bounds(level, cell_iarr)
        yield from range(low_key, high_key)

        # Bottom & Top
        pbox = bbox.copy()
        pbox[0, 0] -= pad[0]
        pbox[0, 1] += pad[0]
        pbox[1, 0] -= pad[1]
        pbox[1, 1] += pad[1]
        pbox[2, 0] -= pad[2]
        pbox[2, 1] = bbox[2, 0]
        yield from self.get_bbox(pbox[:, 0], pbox[:, 1])

        pbox[2, 0] = bbox[2, 1]
        pbox[2, 1] = pbox[2, 0] + pad[2]
        yield from self.get_bbox(pbox[:, 0], pbox[:, 1])

        # Front & Back
        pbox = bbox.copy()
        pbox[0, 0] -= pad[0]
        pbox[0, 1] += pad[0]
        pbox[1, 0] -= pad[1]
        pbox[1, 1] = bbox[1, 0]
        yield from self.get_bbox(pbox[:, 0], pbox[:, 1])
        pbox[1, 0] = bbox[1, 1]
        pbox[1, 1] = pbox[1, 0] + pad[1]
        yield from self.get_bbox(pbox[:, 0], pbox[:, 1])

        # Left & Right
        pbox = bbox.copy()
        pbox[0, 0] -= pad[0]
        pbox[0, 1] = bbox[0, 0]
        yield from self.get_bbox(pbox[:, 0], pbox[:, 1])
        pbox[0, 0] = bbox[0, 1]
        pbox[0, 1] = pbox[0, 0] + pad[0]
        yield from self.get_bbox(pbox[:, 0], pbox[:, 1])



cdef class RegionSelector(SelectorObject):
    cdef public np.float64_t left_edge[3]
    cdef public np.float64_t right_edge[3]
    cdef public np.float64_t right_edge_shift[3]
    cdef public bint is_all_data
    cdef public bint loose_selection
    cdef public bint check_period[3]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __init__(self, dobj):
        cdef int i
        # We are modifying dobj.left_edge and dobj.right_edge , so here we will
        # do an in-place conversion of those arrays.
        cdef np.float64_t[:] RE = _ensure_code(dobj.right_edge)
        cdef np.float64_t[:] LE = _ensure_code(dobj.left_edge)
        cdef np.float64_t[:] DW = _ensure_code(dobj.ds.domain_width)
        cdef np.float64_t[:] DLE = _ensure_code(dobj.ds.domain_left_edge)
        cdef np.float64_t[:] DRE = _ensure_code(dobj.ds.domain_right_edge)
        le_all = (np.array(LE) == _ensure_code(dobj.ds.domain_left_edge)).all()
        re_all = (np.array(RE) == _ensure_code(dobj.ds.domain_right_edge)).all()
        # If we have a bounding box, then we should *not* revert to all data
        domain_override = getattr(dobj.ds, '_domain_override', False)
        if le_all and re_all and not domain_override:
            self.is_all_data = True
        else:
            self.is_all_data = False
        cdef np.float64_t region_width[3]
        cdef bint p[3]
        # This is for if we want to include zones that overlap and whose
        # centers are not strictly included.
        self.loose_selection = getattr(dobj, "loose_selection", False)

        for i in range(3):
            self.check_period[i] = False
            region_width[i] = RE[i] - LE[i]
            p[i] = dobj.ds.periodicity[i]
            if region_width[i] <= 0:
                raise RuntimeError(
                    "Region right edge[%s] < left edge: width = %s" % (
                        i, region_width[i]))

        for i in range(3):

            if p[i]:
                # First, we check if any criteria requires a period check,
                # without any adjustments.  This is for short-circuiting the
                # short-circuit of the loop down below in mask filling.
                if LE[i] < DLE[i] or LE[i] > DRE[i] or RE[i] > DRE[i]:
                    self.check_period[i] = True
                # shift so left_edge guaranteed in domain
                if LE[i] < DLE[i]:
                    LE[i] += DW[i]
                    RE[i] += DW[i]
                elif LE[i] > DRE[i]:
                    LE[i] -= DW[i]
                    RE[i] -= DW[i]
            else:
                if LE[i] < DLE[i] or RE[i] > DRE[i]:
                    raise RuntimeError(
                        "yt attempted to read outside the boundaries of "
                        "a non-periodic domain along dimension %s.\n"
                        "Region left edge = %s, Region right edge = %s\n"
                        "Dataset left edge = %s, Dataset right edge = %s\n\n"
                        "This commonly happens when trying to compute ghost cells "
                        "up to the domain boundary. Two possible solutions are to "
                        "select a smaller region that does not border domain edge "
                        "(see https://yt-project.org/docs/analyzing/objects.html?highlight=region)\n"
                        "or override the periodicity with\n"
                        "ds.force_periodicity()" % \
                        (i, dobj.left_edge[i], dobj.right_edge[i],
                         dobj.ds.domain_left_edge[i], dobj.ds.domain_right_edge[i])
                    )
            # Already ensured in code
            self.left_edge[i] = LE[i]
            self.right_edge[i] = RE[i]
            self.right_edge_shift[i] = RE[i] - DW[i]
            if not self.periodicity[i]:
                self.right_edge_shift[i] = -np.inf

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) nogil:
        cdef int i
        for i in range(3):
            if (right_edge[i] < self.left_edge[i] and \
                left_edge[i] >= self.right_edge_shift[i]) or \
                left_edge[i] >= self.right_edge[i]:
                return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) nogil:
        cdef int i
        for i in range(3):
            if (right_edge[i] < self.left_edge[i] and \
                left_edge[i] >= self.right_edge_shift[i]) or \
                left_edge[i] >= self.right_edge[i]:
                return 0
        for i in range(3):
            if left_edge[i] < self.right_edge_shift[i]:
                if right_edge[i] >= self.right_edge_shift[i]:
                    return 2
            elif left_edge[i] < self.left_edge[i] or \
                 right_edge[i] >= self.right_edge[i]:
                return 2
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) nogil:
        cdef np.float64_t left_edge[3]
        cdef np.float64_t right_edge[3]
        cdef int i
        if self.loose_selection:
            for i in range(3):
                left_edge[i] = pos[i] - dds[i]*0.5
                right_edge[i] = pos[i] + dds[i]*0.5
            return self.select_bbox(left_edge, right_edge)
        return self.select_point(pos)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_point(self, np.float64_t pos[3]) nogil:
        cdef int i
        for i in range(3):
            if (self.right_edge_shift[i] <= pos[i] < self.left_edge[i]) or \
               pos[i] >= self.right_edge[i]:
                return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) nogil:
        # adapted from http://stackoverflow.com/a/4579192/1382869
        cdef int i
        cdef np.float64_t p
        cdef np.float64_t r2 = radius**2
        cdef np.float64_t dmin = 0
        for i in range(3):
            if self.periodicity[i] and self.check_period[i]:
                p = pos[i] + self.right_edge_shift[i]
            else:
                p = pos[i]
            if p < self.left_edge[i]:
                dmin += (p - self.left_edge[i])**2
            elif pos[i] > self.right_edge[i]:
                dmin += (p - self.right_edge[i])**2
        return int(dmin <= r2)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int fill_mask_selector(self, np.float64_t left_edge[3],
                                np.float64_t right_edge[3],
                                np.float64_t dds[3], int dim[3],
                                np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask,
                                np.ndarray[np.uint8_t, ndim=3] mask,
                                int level):
        cdef int i, j, k
        cdef int total = 0, this_level = 0
        cdef np.float64_t pos[3]
        if level < self.min_level or level > self.max_level:
            return 0
        if level == self.max_level:
            this_level = 1
        cdef np.int64_t si[3]
        cdef np.int64_t ei[3]
        for i in range(3):
            if not self.check_period[i]:
                si[i] = <np.int64_t> ((self.left_edge[i] - left_edge[i])/dds[i])
                ei[i] = <np.int64_t> ((self.right_edge[i] - left_edge[i])/dds[i])
                si[i] = iclip(si[i] - 1, 0, dim[i])
                ei[i] = iclip(ei[i] + 1, 0, dim[i])
            else:
                si[i] = 0
                ei[i] = dim[i]
        with nogil:

            for i in range(si[0], ei[0]):
                pos[0] = left_edge[0] + (i + 0.5) * dds[0]
                for j in range(si[1], ei[1]):
                    pos[1] = left_edge[1] + (j + 0.5) * dds[1]
                    for k in range(si[2], ei[2]):
                        pos[2] = left_edge[2] + (k + 0.5) * dds[2]
                        if child_mask[i, j, k] == 1 or this_level == 1:
                            mask[i, j, k] = self.select_cell(pos, dds)
                            total += mask[i, j, k]
        return total

    def _hash_vals(self):
        return (("left_edge[0]", self.left_edge[0]),
                ("left_edge[1]", self.left_edge[1]),
                ("left_edge[2]", self.left_edge[2]),
                ("right_edge[0]", self.right_edge[0]),
                ("right_edge[1]", self.right_edge[1]),
                ("right_edge[2]", self.right_edge[2]))

    def _get_state_attnames(self):
        return ('left_edge', 'right_edge', 'right_edge_shift', 'check_period',
                'is_all_data', 'loose_selection')

region_selector = RegionSelector



cdef class DataCollectionSelector(SelectorObject):
    cdef object obj_ids
    cdef np.int64_t nids

    def __init__(self, dobj):
        self.obj_ids = dobj._obj_ids
        self.nids = self.obj_ids.shape[0]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def select_grids(self,
                     np.ndarray[np.float64_t, ndim=2] left_edges,
                     np.ndarray[np.float64_t, ndim=2] right_edges,
                     np.ndarray[np.int32_t, ndim=2] levels):
        cdef int n
        cdef int ng = left_edges.shape[0]
        cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
        cdef np.ndarray[np.int64_t, ndim=1] oids = self.obj_ids
        with nogil:
            for n in range(self.nids):
                gridi[oids[n]] = 1
        return gridi.astype("bool")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_mask(self, gobj):
        cdef np.ndarray[np.uint8_t, ndim=3] mask
        mask = np.ones(gobj.ActiveDimensions, dtype='uint8')
        return mask.astype("bool")

    def _hash_vals(self):
        return (hash(self.obj_ids.tobytes()), self.nids)

data_collection_selector = DataCollectionSelector



# distutils: include_dirs = LIB_DIR
# distutils: libraries = STD_LIBS
# distutils: sources = FIXED_INTERP
# distutils: language = c++
# distutils: extra_compile_args = CPP14_FLAG
# distutils: extra_link_args = CPP14_FLAG
"""
Simple integrators for the radiative transfer equation



"""


import numpy as np

cimport cython
cimport numpy as np
from fixed_interpolator cimport *
from libc.math cimport (
    M_PI,
    acos,
    asin,
    atan,
    atan2,
    cos,
    exp,
    fabs,
    floor,
    log2,
    sin,
    sqrt,
)

#cimport healpix_interface
from libc.stdlib cimport abs, calloc, free, malloc

from yt.utilities.lib.fp_utils cimport fclip, fmax, fmin, i64clip, iclip, imax, imin

DEF Nch = 4

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef int walk_volume(VolumeContainer *vc,
                     np.float64_t v_pos[3],
                     np.float64_t v_dir[3],
                     sampler_function *sample,
                     void *data,
                     np.float64_t *return_t = NULL,
                     np.float64_t max_t = 1.0) nogil:
    cdef int cur_ind[3]
    cdef int step[3]
    cdef int x, y, i, hit, direction
    cdef np.float64_t intersect_t = 1.1
    cdef np.float64_t iv_dir[3]
    cdef np.float64_t tmax[3]
    cdef np.float64_t tdelta[3]
    cdef np.float64_t exit_t = -1.0, enter_t = -1.0
    cdef np.float64_t tl, temp_x, temp_y = -1
    if max_t > 1.0: max_t = 1.0
    direction = -1
    if vc.left_edge[0] <= v_pos[0] and v_pos[0] < vc.right_edge[0] and \
       vc.left_edge[1] <= v_pos[1] and v_pos[1] < vc.right_edge[1] and \
       vc.left_edge[2] <= v_pos[2] and v_pos[2] < vc.right_edge[2]:
        intersect_t = 0.0
        direction = 3
    for i in range(3):
        if (v_dir[i] < 0):
            step[i] = -1
        elif (v_dir[i] == 0.0):
            step[i] = 0
            continue
        else:
            step[i] = 1
        iv_dir[i] = 1.0/v_dir[i]
        if direction == 3: continue
        x = (i+1) % 3
        y = (i+2) % 3
        if step[i] > 0:
            tl = (vc.left_edge[i] - v_pos[i])*iv_dir[i]
        else:
            tl = (vc.right_edge[i] - v_pos[i])*iv_dir[i]
        temp_x = (v_pos[x] + tl*v_dir[x])
        temp_y = (v_pos[y] + tl*v_dir[y])
        if fabs(temp_x - vc.left_edge[x]) < 1e-10*vc.dds[x]:
            temp_x = vc.left_edge[x]
        elif fabs(temp_x - vc.right_edge[x]) < 1e-10*vc.dds[x]:
            temp_x = vc.right_edge[x]
        if fabs(temp_y - vc.left_edge[y]) < 1e-10*vc.dds[y]:
            temp_y = vc.left_edge[y]
        elif fabs(temp_y - vc.right_edge[y]) < 1e-10*vc.dds[y]:
            temp_y = vc.right_edge[y]
        if vc.left_edge[x] <= temp_x and temp_x <= vc.right_edge[x] and \
           vc.left_edge[y] <= temp_y and temp_y <= vc.right_edge[y] and \
           0.0 <= tl and tl < intersect_t:
            direction = i
            intersect_t = tl
    if enter_t >= 0.0: intersect_t = enter_t
    if not ((0.0 <= intersect_t) and (intersect_t < max_t)): return 0
    for i in range(3):
        # Two things have to be set inside this loop.
        # cur_ind[i], the current index of the grid cell the ray is in
        # tmax[i], the 't' until it crosses out of the grid cell
        tdelta[i] = step[i] * iv_dir[i] * vc.dds[i]
        if i == direction and step[i] > 0:
            # Intersection with the left face in this direction
            cur_ind[i] = 0
        elif i == direction and step[i] < 0:
            # Intersection with the right face in this direction
            cur_ind[i] = vc.dims[i] - 1
        else:
            # We are somewhere in the middle of the face
            temp_x = intersect_t * v_dir[i] + v_pos[i] # current position
            temp_y = ((temp_x - vc.left_edge[i])*vc.idds[i])
            # There are some really tough cases where we just within a couple
            # least significant places of the edge, and this helps prevent
            # killing the calculation through a segfault in those cases.
            if -1 < temp_y < 0 and step[i] > 0:
                temp_y = 0.0
            elif vc.dims[i] - 1 < temp_y < vc.dims[i] and step[i] < 0:
                temp_y = vc.dims[i] - 1
            cur_ind[i] =  <int> (floor(temp_y))
        if step[i] > 0:
            temp_y = (cur_ind[i] + 1) * vc.dds[i] + vc.left_edge[i]
        elif step[i] < 0:
            temp_y = cur_ind[i] * vc.dds[i] + vc.left_edge[i]
        tmax[i] = (temp_y - v_pos[i]) * iv_dir[i]
        if step[i] == 0:
            tmax[i] = 1e60
    # We have to jumpstart our calculation
    for i in range(3):
        if cur_ind[i] == vc.dims[i] and step[i] >= 0:
            return 0
        if cur_ind[i] == -1 and step[i] <= -1:
            return 0
    enter_t = intersect_t
    hit = 0
    while 1:
        hit += 1
        if tmax[0] < tmax[1]:
            if tmax[0] < tmax[2]:
                i = 0
            else:
                i = 2
        else:
            if tmax[1] < tmax[2]:
                i = 1
            else:
                i = 2
        exit_t = fmin(tmax[i], max_t)
        sample(vc, v_pos, v_dir, enter_t, exit_t, cur_ind, data)
        cur_ind[i] += step[i]
        enter_t = tmax[i]
        tmax[i] += tdelta[i]
        if cur_ind[i] < 0 or cur_ind[i] >= vc.dims[i] or enter_t >= max_t:
            break
    if return_t != NULL: return_t[0] = exit_t
    return hit

def hp_pix2vec_nest(long nside, long ipix):
    raise NotImplementedError
    # cdef double v[3]
    # healpix_interface.pix2vec_nest(nside, ipix, v)
    # cdef np.ndarray[np.float64_t, ndim=1] tr = np.empty((3,), dtype='float64')
    # tr[0] = v[0]
    # tr[1] = v[1]
    # tr[2] = v[2]
    # return tr

def arr_pix2vec_nest(long nside,
                     np.ndarray[np.int64_t, ndim=1] aipix):
    raise NotImplementedError
    # cdef int n = aipix.shape[0]
    # cdef int i
    # cdef double v[3]
    # cdef long ipix
    # cdef np.ndarray[np.float64_t, ndim=2] tr = np.zeros((n, 3), dtype='float64')
    # for i in range(n):
    #     ipix = aipix[i]
    #     healpix_interface.pix2vec_nest(nside, ipix, v)
    #     tr[i,0] = v[0]
    #     tr[i,1] = v[1]
    #     tr[i,2] = v[2]
    # return tr

def hp_vec2pix_nest(long nside, double x, double y, double z):
    raise NotImplementedError
    # cdef double v[3]
    # v[0] = x
    # v[1] = y
    # v[2] = z
    # cdef long ipix
    # healpix_interface.vec2pix_nest(nside, v, &ipix)
    # return ipix

def arr_vec2pix_nest(long nside,
                     np.ndarray[np.float64_t, ndim=1] x,
                     np.ndarray[np.float64_t, ndim=1] y,
                     np.ndarray[np.float64_t, ndim=1] z):
    raise NotImplementedError
    # cdef int n = x.shape[0]
    # cdef int i
    # cdef double v[3]
    # cdef long ipix
    # cdef np.ndarray[np.int64_t, ndim=1] tr = np.zeros(n, dtype='int64')
    # for i in range(n):
    #     v[0] = x[i]
    #     v[1] = y[i]
    #     v[2] = z[i]
    #     healpix_interface.vec2pix_nest(nside, v, &ipix)
    #     tr[i] = ipix
    # return tr

def hp_pix2ang_nest(long nside, long ipnest):
    raise NotImplementedError
    # cdef double theta, phi
    # healpix_interface.pix2ang_nest(nside, ipnest, &theta, &phi)
    # return (theta, phi)

def arr_pix2ang_nest(long nside, np.ndarray[np.int64_t, ndim=1] aipnest):
    raise NotImplementedError
    # cdef int n = aipnest.shape[0]
    # cdef int i
    # cdef long ipnest
    # cdef np.ndarray[np.float64_t, ndim=2] tr = np.zeros((n, 2), dtype='float64')
    # cdef double theta, phi
    # for i in range(n):
    #     ipnest = aipnest[i]
    #     healpix_interface.pix2ang_nest(nside, ipnest, &theta, &phi)
    #     tr[i,0] = theta
    #     tr[i,1] = phi
    # return tr

def hp_ang2pix_nest(long nside, double theta, double phi):
    raise NotImplementedError
    # cdef long ipix
    # healpix_interface.ang2pix_nest(nside, theta, phi, &ipix)
    # return ipix

def arr_ang2pix_nest(long nside,
                     np.ndarray[np.float64_t, ndim=1] atheta,
                     np.ndarray[np.float64_t, ndim=1] aphi):
    raise NotImplementedError
    # cdef int n = atheta.shape[0]
    # cdef int i
    # cdef long ipnest
    # cdef np.ndarray[np.int64_t, ndim=1] tr = np.zeros(n, dtype='int64')
    # cdef double theta, phi
    # for i in range(n):
    #     theta = atheta[i]
    #     phi = aphi[i]
    #     healpix_interface.ang2pix_nest(nside, theta, phi, &ipnest)
    #     tr[i] = ipnest
    # return tr

@cython.boundscheck(False)
@cython.cdivision(False)
@cython.wraparound(False)
def pixelize_healpix(long nside,
                     np.ndarray[np.float64_t, ndim=1] values,
                     long ntheta, long nphi,
                     np.ndarray[np.float64_t, ndim=2] irotation):
    raise NotImplementedError
    # # We will first to pix2vec, rotate, then calculate the angle
    # cdef int i, j, thetai, phii
    # cdef long ipix
    # cdef double v0[3], v1[3]
    # cdef double pi = 3.1415926
    # cdef np.float64_t pi2 = pi/2.0
    # cdef np.float64_t phi, theta
    # cdef np.ndarray[np.float64_t, ndim=2] results
    # cdef np.ndarray[np.int32_t, ndim=2] count
    # results = np.zeros((ntheta, nphi), dtype="float64")
    # count = np.zeros((ntheta, nphi), dtype="int32")

    # cdef np.float64_t phi0 = 0
    # cdef np.float64_t dphi = 2.0 * pi/(nphi-1)

    # cdef np.float64_t theta0 = 0
    # cdef np.float64_t dtheta = pi/(ntheta-1)
    # # We assume these are the rotated theta and phi
    # for thetai in range(ntheta):
    #     theta = theta0 + dtheta * thetai
    #     for phii in range(nphi):
    #         phi = phi0 + dphi * phii
    #         # We have our rotated vector
    #         v1[0] = cos(phi) * sin(theta)
    #         v1[1] = sin(phi) * sin(theta)
    #         v1[2] = cos(theta)
    #         # Now we rotate back
    #         for i in range(3):
    #             v0[i] = 0
    #             for j in range(3):
    #                 v0[i] += v1[j] * irotation[j,i]
    #         # Get the pixel this vector is inside
    #         healpix_interface.vec2pix_nest(nside, v0, &ipix)
    #         results[thetai, phii] = values[ipix]
    #         count[i, j] += 1
    # return results, count

def healpix_aitoff_proj(np.ndarray[np.float64_t, ndim=1] pix_image,
                        long nside,
                        np.ndarray[np.float64_t, ndim=2] image,
                        np.ndarray[np.float64_t, ndim=2] irotation):
    raise NotImplementedError
    # cdef double pi = np.pi
    # cdef int i, j, k, l
    # cdef np.float64_t x, y, z, zb
    # cdef np.float64_t dx, dy, inside
    # cdef double v0[3], v1[3]
    # dx = 2.0 / (image.shape[1] - 1)
    # dy = 2.0 / (image.shape[0] - 1)
    # cdef np.float64_t s2 = sqrt(2.0)
    # cdef long ipix
    # for i in range(image.shape[1]):
    #     x = (-1.0 + i*dx)*s2*2.0
    #     for j in range(image.shape[0]):
    #         y = (-1.0 + j * dy)*s2
    #         zb = (x*x/8.0 + y*y/2.0 - 1.0)
    #         if zb > 0: continue
    #         z = (1.0 - (x/4.0)**2.0 - (y/2.0)**2.0)
    #         z = sqrt(z)
    #         # Longitude
    #         phi = (2.0*atan(z*x/(2.0 * (2.0*z*z-1.0))) + pi)
    #         # Latitude
    #         # We shift it into co-latitude
    #         theta = (asin(z*y) + pi/2.0)
    #         # Now to account for rotation we translate into vectors
    #         v1[0] = cos(phi) * sin(theta)
    #         v1[1] = sin(phi) * sin(theta)
    #         v1[2] = cos(theta)
    #         for k in range(3):
    #             v0[k] = 0
    #             for l in range(3):
    #                 v0[k] += v1[l] * irotation[l,k]
    #         healpix_interface.vec2pix_nest(nside, v0, &ipix)
    #         #print("Rotated", v0[0], v0[1], v0[2], v1[0], v1[1], v1[2], ipix, pix_image[ipix])
    #         image[j, i] = pix_image[ipix]

def arr_fisheye_vectors(int resolution, np.float64_t fov, int nimx=1, int
        nimy=1, int nimi=0, int nimj=0, np.float64_t off_theta=0.0, np.float64_t
        off_phi=0.0):
    # We now follow figures 4-7 of:
    # http://paulbourke.net/miscellaneous/domefisheye/fisheye/
    # ...but all in Cython.
    cdef np.ndarray[np.float64_t, ndim=3] vp
    cdef int i, j
    cdef np.float64_t r, phi, theta, px, py
    cdef np.float64_t fov_rad = fov * np.pi / 180.0
    cdef int nx = resolution/nimx
    cdef int ny = resolution/nimy
    vp = np.zeros((nx,ny, 3), dtype="float64")
    for i in range(nx):
        px = (2.0 * (nimi*nx + i)) / resolution - 1.0
        for j in range(ny):
            py = (2.0 * (nimj*ny + j)) / resolution - 1.0
            r = sqrt(px*px + py*py)
            if r > 1.01:
                vp[i,j,0] = vp[i,j,1] = vp[i,j,2] = 0.0
                continue
            phi = atan2(py, px)
            theta = r * fov_rad / 2.0
            theta += off_theta
            phi += off_phi
            vp[i,j,0] = sin(theta) * cos(phi)
            vp[i,j,1] = sin(theta) * sin(phi)
            vp[i,j,2] = cos(theta)
    return vp



from yt.testing import requires_file
from yt.utilities.answer_testing.framework import data_dir_load
from yt.visualization.plot_window import SlicePlot


@requires_file("amrvac/bw_3d0000.dat")
def test_code_units_xy_labels():
    ds = data_dir_load("amrvac/bw_3d0000.dat", kwargs=dict(unit_system="code"))
    p = SlicePlot(ds, "x", ("gas", "density"))

    ax = p.plots[("gas", "density")].axes
    assert "code length" in ax.get_xlabel().replace("\\", "")
    assert "code length" in ax.get_ylabel().replace("\\", "")



"""
AMRVAC-specific IO functions



"""
import os

import numpy as np
from more_itertools import always_iterable

from yt.geometry.selection_routines import GridSelector
from yt.utilities.io_handler import BaseIOHandler
from yt.utilities.on_demand_imports import _f90nml as f90nml

from .datfile_utils import get_single_block_field_data


def read_amrvac_namelist(parfiles):
    """Read one or more parfiles, and return a unified f90nml.Namelist object.

    This function replicates the patching logic of MPI-AMRVAC where redundant parameters
    only retain last-in-line values, with the exception of `&filelist:base_filename`,
    which is accumulated. When passed a single file, this function acts as a mere
    wrapper of f90nml.read().

    Parameters
    ----------
    parfiles : str, os.Pathlike, byte, or an iterable returning those types
        A file path, or a list of file paths to MPI-AMRVAC configuration parfiles.

    Returns
    -------
    unified_namelist : f90nml.Namelist
        A single namelist object. The class inherits from ordereddict.

    """
    parfiles = (os.path.expanduser(pf) for pf in always_iterable(parfiles))

    # first merge the namelists
    namelists = [f90nml.read(parfile) for parfile in parfiles]
    unified_namelist = f90nml.Namelist()
    for nml in namelists:
        unified_namelist.patch(nml)

    # accumulate `&filelist:base_filename`
    base_filename = "".join(
        nml.get("filelist", {}).get("base_filename", "") for nml in namelists
    )
    unified_namelist["filelist"]["base_filename"] = base_filename

    return unified_namelist


class AMRVACIOHandler(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "amrvac"

    def __init__(self, ds):
        BaseIOHandler.__init__(self, ds)
        self.ds = ds
        self.datfile = ds.parameter_filename
        header = self.ds.parameters
        self.block_shape = np.append(header["block_nx"], header["nw"])

    def _read_particle_coords(self, chunks, ptf):
        """Not implemented yet."""
        # This needs to *yield* a series of tuples of (ptype, (x, y, z)).
        # chunks is a list of chunks, and ptf is a dict where the keys are
        # ptypes and the values are lists of fields.
        raise NotImplementedError

    def _read_particle_fields(self, chunks, ptf, selector):
        """Not implemented yet."""
        # This gets called after the arrays have been allocated.  It needs to
        # yield ((ptype, field), data) where data is the masked results of
        # reading ptype, field and applying the selector to the data read in.
        # Selector objects have a .select_points(x,y,z) that returns a mask, so
        # you need to do your masking here.
        raise NotImplementedError

    def _read_data(self, grid, field):
        """Retrieve field data from a grid.

        Parameters
        ----------
        grid : yt.frontends.amrvac.data_structures.AMRVACGrid
            The grid from which data is to be read.
        field : str
            A field name.

        Returns
        -------
        data : np.ndarray
            A 3D array of float64 type representing grid data.

        """
        ileaf = grid.id
        offset = grid._index.block_offsets[ileaf]
        field_idx = self.ds.parameters["w_names"].index(field)
        with open(self.datfile, "rb") as istream:
            data = get_single_block_field_data(
                istream, offset, self.block_shape, field_idx
            )

        # Always convert data to 3D, as grid.ActiveDimensions is always 3D
        while len(data.shape) < 3:
            data = data[..., np.newaxis]
        return data

    def _read_fluid_selection(self, chunks, selector, fields, size):
        """Retrieve field(s) data in a selected region of space.

        Parameters
        ----------
        chunks : generator
            A generator for multiple chunks, each of which contains a list of grids.

        selector : yt.geometry.selection_routines.SelectorObject
            A spatial region selector.

        fields : list
            A list of tuples (ftype, fname).

        size : np.int64
            The cumulative number of objs contained in all chunks.

        Returns
        -------
        data_dict : dict
            keys are the (ftype, fname) tuples, values are arrays that have been masked
            using whatever selector method is appropriate. Arrays have dtype float64.
        """

        # @Notes from Niels:
        # The chunks list has YTDataChunk objects containing the different grids.
        # The list of grids can be obtained by doing eg.
        # grids_list = chunks[0].objs or chunks[1].objs etc.
        # Every element in "grids_list" is then an AMRVACGrid object,
        # and has hence all attributes of a grid :
        # (Level, ActiveDimensions, LeftEdge, etc.)

        chunks = list(chunks)
        data_dict = {}  # <- return variable
        if isinstance(selector, GridSelector):
            if not len(chunks) == len(chunks[0].objs) == 1:
                raise RuntimeError
            grid = chunks[0].objs[0]
            for ftype, fname in fields:
                data_dict[ftype, fname] = self._read_data(grid, fname)
        else:
            if size is None:
                size = sum(g.count(selector) for chunk in chunks for g in chunk.objs)

            for field in fields:
                data_dict[field] = np.empty(size, dtype="float64")

            # nb_grids = sum(len(chunk.objs) for chunk in chunks)

            ind = 0
            for chunk in chunks:
                for grid in chunk.objs:
                    nd = 0
                    for field in fields:
                        ftype, fname = field
                        data = self._read_data(grid, fname)
                        nd = grid.select(selector, data, data_dict[field], ind)
                    ind += nd

        return data_dict

    def _read_chunk_data(self, chunk, fields):
        """Not implemented yet."""
        # This reads the data from a single chunk without doing any selection,
        # and is only used for caching data that might be used by multiple
        # different selectors later. For instance, this can speed up ghost zone
        # computation.
        # it should be used by _read_fluid_selection instead of _read_data
        raise NotImplementedError



from .camera import Camera
from .image_handling import export_rgba, import_rgba, plot_channel, plot_rgb
from .off_axis_projection import off_axis_projection
from .render_source import (
    BoxSource,
    CoordinateVectorSource,
    GridSource,
    LineSource,
    MeshSource,
    OpaqueSource,
    PointSource,
    create_volume_source,
)
from .scene import Scene
from .transfer_function_helper import TransferFunctionHelper
from .transfer_functions import (
    ColorTransferFunction,
    MultiVariateTransferFunction,
    PlanckTransferFunction,
    ProjectionTransferFunction,
    TransferFunction,
)
from .volume_rendering import create_scene, volume_render
from .zbuffer_array import ZBuffer



"""
AdaptaHOP data-file handling function




"""


from operator import attrgetter

import numpy as np

from yt.utilities.cython_fortran_utils import FortranFile
from yt.utilities.io_handler import BaseIOHandler

from .definitions import HALO_ATTRIBUTES, HEADER_ATTRIBUTES


class IOHandlerAdaptaHOPBinary(BaseIOHandler):
    _dataset_type = "adaptahop_binary"

    _offsets = None  # Location of halos in the file
    _particle_positions = None  # Buffer of halo position

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _read_chunk_data(self, chunk, fields):
        raise NotImplementedError

    def _yield_coordinates(self, data_file):
        yield "halos", self._get_particle_positions()

    def _read_particle_coords(self, chunks, ptf):
        # This will read chunks and yield the results.
        chunks = list(chunks)
        data_files = set()
        # Only support halo reading for now.
        assert len(ptf) == 1
        assert list(ptf.keys())[0] == "halos"
        ptype = "halos"
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files, key=attrgetter("filename")):
            pcount = (
                data_file.ds.parameters["nhalos"] + data_file.ds.parameters["nsubs"]
            )
            if pcount == 0:
                continue
            pos = self._get_particle_positions()
            yield ptype, [pos[:, i] for i in range(3)]

    def _read_particle_fields(self, chunks, ptf, selector):
        # Now we have all the sizes, and we can allocate
        chunks = list(chunks)
        data_files = set()
        # Only support halo reading for now.
        assert len(ptf) == 1
        assert list(ptf.keys())[0] == "halos"
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)

        def iterate_over_attributes(attr_list):
            for attr, *_ in attr_list:
                if isinstance(attr, tuple):
                    yield from attr
                else:
                    yield attr

        for data_file in sorted(data_files, key=attrgetter("filename")):
            pcount = (
                data_file.ds.parameters["nhalos"] + data_file.ds.parameters["nsubs"]
            )
            if pcount == 0:
                continue
            ptype = "halos"
            field_list0 = sorted(ptf[ptype], key=_find_attr_position)
            field_list_pos = [f"raw_position_{k}" for k in "xyz"]
            field_list = sorted(
                set(field_list0 + field_list_pos), key=_find_attr_position
            )

            with FortranFile(self.ds.parameter_filename) as fpu:
                params = fpu.read_attrs(HEADER_ATTRIBUTES)

                todo = _todo_from_attributes(field_list)

                nhalos = params["nhalos"] + params["nsubs"]
                data = np.zeros((nhalos, len(field_list)))
                for ihalo in range(nhalos):
                    jj = 0
                    for it in todo:
                        if isinstance(it, int):
                            fpu.skip(it)
                        else:
                            tmp = fpu.read_attrs(it)
                            for key in iterate_over_attributes(it):
                                v = tmp[key]
                                if key not in field_list:
                                    continue
                                data[ihalo, jj] = v
                                jj += 1
            ipos = [field_list.index(k) for k in field_list_pos]
            w = self.ds.domain_width.to("code_length")[0].value / 2
            x, y, z = (data[:, i] + w for i in ipos)
            mask = selector.select_points(x, y, z, 0.0)
            del x, y, z

            if mask is None:
                continue
            for field in field_list0:
                i = field_list.index(field)
                yield (ptype, field), data[mask, i]

    def _count_particles(self, data_file):
        nhalos = data_file.ds.parameters["nhalos"] + data_file.ds.parameters["nsubs"]
        return {"halos": nhalos}

    def _identify_fields(self, data_file):
        fields = []
        for attr, _1, _2 in HALO_ATTRIBUTES:
            if isinstance(attr, str):
                fields.append(("halos", attr))
            else:
                for a in attr:
                    fields.append(("halos", a))
        return fields, {}

    # -----------------------------------------------------
    # Specific to AdaptaHOP
    def _get_particle_positions(self):
        """Read the particles and return them in code_units"""
        data = getattr(self, "_particle_positions", None)
        if data is not None:
            return data

        with FortranFile(self.ds.parameter_filename) as fpu:
            params = fpu.read_attrs(HEADER_ATTRIBUTES)

            todo = _todo_from_attributes(
                (
                    "particle_identifier",
                    "raw_position_x",
                    "raw_position_y",
                    "raw_position_z",
                )
            )

            nhalos = params["nhalos"] + params["nsubs"]
            data = np.zeros((nhalos, 3))
            offset_map = np.zeros((nhalos, 2), dtype="int64")
            for ihalo in range(nhalos):
                ipos = fpu.tell()
                for it in todo:
                    if isinstance(it, int):
                        fpu.skip(it)
                    elif it[0][0] != "particle_identifier":
                        # Small optimisation here: we can read as vector
                        # dt = fpu.read_attrs(it)
                        # data[ihalo, 0] = dt['particle_position_x']
                        # data[ihalo, 1] = dt['particle_position_y']
                        # data[ihalo, 2] = dt['particle_position_z']
                        data[ihalo, :] = fpu.read_vector(it[0][-1])
                    else:
                        halo_id = fpu.read_int()
                        offset_map[ihalo, 0] = halo_id
                        offset_map[ihalo, 1] = ipos
        data = self.ds.arr(data, "code_length") + self.ds.domain_width / 2

        # Make sure halos are loaded in increasing halo_id order
        assert np.all(np.diff(offset_map[:, 0]) > 0)

        # Cache particle positions as one do not expect a large number of halos anyway
        self._particle_positions = data
        self._offsets = offset_map
        return data

    def members(self, ihalo):
        offset = self._offsets[ihalo, 1]
        todo = _todo_from_attributes(("particle_identities",))
        with FortranFile(self.ds.parameter_filename) as fpu:
            fpu.seek(offset)
            if isinstance(todo[0], int):
                fpu.skip(todo.pop(0))
            members = fpu.read_attrs(todo.pop(0))["particle_identities"]
        return members


def _todo_from_attributes(attributes):
    # Helper function to generate a list of read-skip instructions given a list of
    # attributes. This is used to skip fields most of the fields when reading
    # the tree_brick files.
    iskip = 0
    todo = []

    attributes = set(attributes)

    for i, (attrs, l, k) in enumerate(HALO_ATTRIBUTES):
        if not isinstance(attrs, tuple):
            attrs_list = (attrs,)
        else:
            attrs_list = attrs
        ok = False
        for attr in attrs_list:
            if attr in attributes:
                ok = True
                break

        if i == 0:
            if ok:
                state = "read"
                todo.append([])
            else:
                state = "skip"

        if ok:
            if state == "skip":
                # Switched from skip to read, store skip information and start
                # new read list
                todo.append(iskip)
                todo.append([])
                iskip = 0
            todo[-1].append((attrs, l, k))
            state = "read"
        else:
            iskip += 1
            state = "skip"

    if state == "skip" and iskip > 0:
        todo.append(iskip)

    return todo


def _find_attr_position(key):
    j = 0
    for attrs, *_ in HALO_ATTRIBUTES:
        if not isinstance(attrs, tuple):
            attrs = (attrs,)
        for a in attrs:
            if key == a:
                return j
            j += 1
    raise KeyError



from yt.frontends.athena.api import AthenaDataset
from yt.loaders import load
from yt.testing import (
    assert_allclose_units,
    assert_equal,
    disable_dataset_cache,
    requires_file,
)
from yt.utilities.answer_testing.framework import (
    data_dir_load,
    requires_ds,
    small_patch_amr,
)

_fields_cloud = (("athena", "scalar[0]"), ("gas", "density"), ("gas", "total_energy"))

cloud = "ShockCloud/id0/Cloud.0050.vtk"


@requires_ds(cloud)
def test_cloud():
    ds = data_dir_load(cloud)
    assert_equal(str(ds), "Cloud.0050")
    for test in small_patch_amr(ds, _fields_cloud):
        test_cloud.__name__ = test.description
        yield test


_fields_blast = (
    ("gas", "temperature"),
    ("gas", "density"),
    ("gas", "velocity_magnitude"),
)

blast = "MHDBlast/id0/Blast.0100.vtk"


@requires_ds(blast)
def test_blast():
    ds = data_dir_load(blast)
    assert_equal(str(ds), "Blast.0100")
    for test in small_patch_amr(ds, _fields_blast):
        test_blast.__name__ = test.description
        yield test


uo_blast = {
    "length_unit": (1.0, "pc"),
    "mass_unit": (2.38858753789e-24, "g/cm**3*pc**3"),
    "time_unit": (1.0, "s*pc/km"),
}


@requires_file(blast)
def test_blast_override():
    # verify that overriding units causes derived unit values to be updated.
    # see issue #1259
    ds = load(blast, units_override=uo_blast)
    assert_equal(float(ds.magnetic_unit.in_units("gauss")), 5.47867467969813e-07)


uo_stripping = {
    "time_unit": 3.086e14,
    "length_unit": 8.0236e22,
    "mass_unit": 9.999e-30 * 8.0236e22 ** 3,
}

_fields_stripping = (
    ("gas", "temperature"),
    ("gas", "density"),
    ("athena", "specific_scalar[0]"),
)

stripping = "RamPressureStripping/id0/rps.0062.vtk"


@requires_ds(stripping, big_data=True)
def test_stripping():
    ds = data_dir_load(stripping, kwargs={"units_override": uo_stripping})
    assert_equal(str(ds), "rps.0062")
    for test in small_patch_amr(ds, _fields_stripping):
        test_stripping.__name__ = test.description
        yield test


sloshing = "MHDSloshing/virgo_low_res.0054.vtk"

uo_sloshing = {
    "length_unit": (1.0, "Mpc"),
    "time_unit": (1.0, "Myr"),
    "mass_unit": (1.0e14, "Msun"),
}


@requires_file(sloshing)
@disable_dataset_cache
def test_nprocs():
    ds1 = load(sloshing, units_override=uo_sloshing)
    sp1 = ds1.sphere("c", (100.0, "kpc"))
    prj1 = ds1.proj(("gas", "density"), 0)
    ds2 = load(sloshing, units_override=uo_sloshing, nprocs=8)
    sp2 = ds2.sphere("c", (100.0, "kpc"))
    prj2 = ds1.proj(("gas", "density"), 0)

    assert_equal(
        sp1.quantities.extrema(("gas", "pressure")),
        sp2.quantities.extrema(("gas", "pressure")),
    )
    assert_allclose_units(
        sp1.quantities.total_quantity(("gas", "pressure")),
        sp2.quantities.total_quantity(("gas", "pressure")),
    )
    for ax in "xyz":
        assert_equal(
            sp1.quantities.extrema(("gas", f"velocity_{ax}")),
            sp2.quantities.extrema(("gas", f"velocity_{ax}")),
        )
    assert_allclose_units(
        sp1.quantities.bulk_velocity(), sp2.quantities.bulk_velocity()
    )
    assert_equal(prj1[("gas", "density")], prj2[("gas", "density")])


@requires_file(cloud)
def test_AthenaDataset():
    assert isinstance(data_dir_load(cloud), AthenaDataset)



import logging
import sys

from yt.config import ytcfg

# This next bit is grabbed from:
# http://stackoverflow.com/questions/384076/how-can-i-make-the-python-logging-output-to-be-colored


def add_coloring_to_emit_ansi(fn):
    # add methods we need to the class
    def new(*args):
        levelno = args[0].levelno
        if levelno >= 50:
            color = "\x1b[31m"  # red
        elif levelno >= 40:
            color = "\x1b[31m"  # red
        elif levelno >= 30:
            color = "\x1b[33m"  # yellow
        elif levelno >= 20:
            color = "\x1b[32m"  # green
        elif levelno >= 10:
            color = "\x1b[35m"  # pink
        else:
            color = "\x1b[0m"  # normal
        ln = color + args[0].levelname + "\x1b[0m"
        args[0].levelname = ln
        return fn(*args)

    return new


def set_log_level(level):
    """
    Select which minimal logging level should be displayed.

    Parameters
    ----------
    level: int or str
        Possible values by increasing level:
        0 or "notset"
        1 or "all"
        10 or "debug"
        20 or "info"
        30 or "warning"
        40 or "error"
        50 or "critical"
    """
    # this is a user-facing interface to avoid importing from yt.utilities in user code.

    if isinstance(level, str):
        level = level.upper()

    if level == "ALL":  # non-standard alias
        level = 1
    ytLogger.setLevel(level)
    ytLogger.debug("Set log level to %s", level)


ufstring = "%(name)-3s: [%(levelname)-9s] %(asctime)s %(message)s"
cfstring = "%(name)-3s: [%(levelname)-18s] %(asctime)s %(message)s"

if ytcfg.get("yt", "stdout_stream_logging"):
    stream = sys.stdout
else:
    stream = sys.stderr

ytLogger = logging.getLogger("yt")


class DuplicateFilter(logging.Filter):
    """A filter that removes duplicated successive log entries."""

    # source
    # https://stackoverflow.com/questions/44691558/suppress-multiple-messages-with-same-content-in-python-logging-module-aka-log-co  # noqa
    def filter(self, record):
        current_log = (record.module, record.levelno, record.msg, record.args)
        if current_log != getattr(self, "last_log", None):
            self.last_log = current_log
            return True
        return False


ytLogger.addFilter(DuplicateFilter())


class DeprecatedFieldFilter(logging.Filter):
    """A filter that suppresses repeated logging of deprecated field warnings"""

    def __init__(self, name=""):
        self.logged_fields = []
        super().__init__(name=name)

    def filter(self, record):
        if not record.msg.startswith("The Derived Field"):
            return True

        field = record.args[0]
        if field in self.logged_fields:
            return False

        self.logged_fields.append(field)
        return True


ytLogger.addFilter(DeprecatedFieldFilter())


def disable_stream_logging():
    if len(ytLogger.handlers) > 0:
        ytLogger.removeHandler(ytLogger.handlers[0])
    h = logging.NullHandler()
    ytLogger.addHandler(h)


def colorize_logging():
    f = logging.Formatter(cfstring)
    ytLogger.handlers[0].setFormatter(f)
    yt_sh.emit = add_coloring_to_emit_ansi(yt_sh.emit)


def uncolorize_logging():
    try:
        f = logging.Formatter(ufstring)
        ytLogger.handlers[0].setFormatter(f)
        yt_sh.emit = original_emitter
    except NameError:
        # yt_sh and original_emitter are not defined because
        # suppress_stream_logging is True, so we continue since there is nothing
        # to uncolorize
        pass


_level = min(max(ytcfg.get("yt", "log_level"), 0), 50)

if ytcfg.get("yt", "suppress_stream_logging"):
    disable_stream_logging()
else:
    yt_sh = logging.StreamHandler(stream=stream)
    # create formatter and add it to the handlers
    formatter = logging.Formatter(ufstring)
    yt_sh.setFormatter(formatter)
    # add the handler to the logger
    ytLogger.addHandler(yt_sh)
    set_log_level(_level)
    ytLogger.propagate = False

    original_emitter = yt_sh.emit

    if ytcfg.get("yt", "colored_logs"):
        colorize_logging()



"""
yt is a toolkit for analyzing and visualizing volumetric data.

* Website: https://yt-project.org
* Documentation: https://yt-project.org/doc
* Data hub: https://girder.hub.yt
* Contribute: https://github.com/yt-project/yt

"""
__version__ = "4.0.1"

import yt.units as units
import yt.utilities.physical_constants as physical_constants
from yt.data_objects.api import (
    DatasetSeries,
    ImageArray,
    ParticleProfile,
    Profile1D,
    Profile2D,
    Profile3D,
    add_particle_filter,
    create_profile,
    particle_filter,
)
from yt.fields.api import (
    DerivedField,
    FieldDetector,
    FieldInfoContainer,
    ValidateDataField,
    ValidateGridType,
    ValidateParameter,
    ValidateProperty,
    ValidateSpatial,
    add_field,
    add_xray_emissivity_field,
    derived_field,
    field_plugins,
)
from yt.frontends.api import _frontend_container
from yt.funcs import (
    enable_plugins,
    get_memory_usage,
    get_pbar,
    get_version_stack,
    get_yt_version,
    insert_ipython,
    is_root,
    is_sequence,
    memory_checker,
    only_on_root,
    parallel_profile,
    print_tb,
    rootonly,
    toggle_interactivity,
)
from yt.units import (
    YTArray,
    YTQuantity,
    display_ytarray,
    loadtxt,
    savetxt,
    uconcatenate,
    ucross,
    udot,
    uhstack,
    uintersect1d,
    unorm,
    ustack,
    uunion1d,
    uvstack,
)
from yt.units.unit_object import define_unit
from yt.utilities.logger import set_log_level, ytLogger as mylog

frontends = _frontend_container()

import yt.visualization.volume_rendering.api as volume_rendering
from yt.frontends.stream.api import hexahedral_connectivity
from yt.frontends.ytdata.api import save_as_dataset
from yt.loaders import simulation  # deprecated alias for load_simulation
from yt.loaders import (
    load,
    load_amr_grids,
    load_hexahedral_mesh,
    load_octree,
    load_particles,
    load_sample,
    load_simulation,
    load_uniform_grid,
    load_unstructured_mesh,
)
from yt.testing import run_nose
from yt.units.unit_systems import UnitSystem, unit_system_registry

# Import some helpful math utilities
from yt.utilities.math_utils import ortho_find, periodic_position, quartiles
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    communication_system,
    enable_parallelism,
    parallel_objects,
)

# Now individual component imports from the visualization API
from yt.visualization.api import (
    AxisAlignedSlicePlot,
    FITSImageData,
    FITSOffAxisProjection,
    FITSOffAxisSlice,
    FITSParticleProjection,
    FITSProjection,
    FITSSlice,
    FixedResolutionBuffer,
    LineBuffer,
    LinePlot,
    OffAxisProjectionPlot,
    OffAxisSlicePlot,
    ParticleImageBuffer,
    ParticlePhasePlot,
    ParticlePlot,
    ParticleProjectionPlot,
    PhasePlot,
    ProfilePlot,
    ProjectionPlot,
    SlicePlot,
    add_colormap,
    apply_colormap,
    make_colormap,
    plot_2d,
    scale_image,
    show_colormaps,
    write_bitmap,
    write_image,
    write_projection,
)
from yt.visualization.volume_rendering.api import (
    ColorTransferFunction,
    TransferFunction,
    create_scene,
    off_axis_projection,
    volume_render,
)

#    TransferFunctionHelper, MultiVariateTransferFunction
#    off_axis_projection


def _check_deprecated_parameters():
    from yt._maintenance.deprecation import issue_deprecation_warning
    from yt.config import ytcfg

    if ytcfg.get("yt", "load_field_plugins"):
        issue_deprecation_warning(
            "Found deprecated parameter 'load_field_plugins' in yt's configuration file.",
            removal="4.1.0",
        )


_check_deprecated_parameters()



# distutils: include_dirs = LIB_DIR
# distutils: extra_compile_args = CPP14_FLAG OMP_ARGS
# distutils: extra_link_args = CPP14_FLAG OMP_ARGS
# distutils: libraries = STD_LIBS
# distutils: sources = FIXED_INTERP
# distutils: language = c++
"""
Image sampler definitions



"""


import numpy as np

cimport cython
cimport lenses
from field_interpolation_tables cimport (
    FieldInterpolationTable,
    FIT_eval_transfer,
    FIT_eval_transfer_with_light,
    FIT_initialize_table,
)
from libc.math cimport sqrt
from libc.stdlib cimport free, malloc

from yt.utilities.lib.fp_utils cimport fclip, i64clip, imin

from .fixed_interpolator cimport (
    eval_gradient,
    fast_interpolate,
    offset_fill,
    offset_interpolate,
    trilinear_interpolate,
    vertex_interp,
)
from .grid_traversal cimport sampler_function, walk_volume

from yt.funcs import mylog

from ._octree_raytracing cimport RayInfo, _OctreeRayTracing


cdef extern from "platform_dep.h":
    long int lrint(double x) nogil

DEF Nch = 4

from cython.parallel import parallel, prange, threadid

from cpython.exc cimport PyErr_CheckSignals
from vec3_ops cimport L2_norm, dot, fma, subtract


cdef struct VolumeRenderAccumulator:
    int n_fits
    int n_samples
    FieldInterpolationTable *fits
    int field_table_ids[6]
    np.float64_t star_coeff
    np.float64_t star_er
    np.float64_t star_sigma_num
    np.float64_t *light_dir
    np.float64_t *light_rgba
    int grey_opacity


cdef class ImageSampler:
    def __init__(self,
                  np.float64_t[:,:,:] vp_pos,
                  np.float64_t[:,:,:] vp_dir,
                  np.ndarray[np.float64_t, ndim=1] center,
                  bounds,
                  np.ndarray[np.float64_t, ndim=3] image,
                  np.ndarray[np.float64_t, ndim=1] x_vec,
                  np.ndarray[np.float64_t, ndim=1] y_vec,
                  np.ndarray[np.float64_t, ndim=1] width,
                  str volume_method,
                  *args,
                  **kwargs):
        cdef int i

        self.volume_method = volume_method
        camera_data = kwargs.pop("camera_data", None)
        if camera_data is not None:
            self.camera_data = camera_data

        zbuffer = kwargs.pop("zbuffer", None)
        if zbuffer is None:
            zbuffer = np.ones((image.shape[0], image.shape[1]), "float64")

        image_used = np.zeros((image.shape[0], image.shape[1]), "int64")
        mesh_lines = np.zeros((image.shape[0], image.shape[1]), "int64")

        self.lens_type = kwargs.pop("lens_type", None)
        if self.lens_type == "plane-parallel":
            self.extent_function = lenses.calculate_extent_plane_parallel
            self.vector_function = lenses.generate_vector_info_plane_parallel
        else:
            if not (vp_pos.shape[0] == vp_dir.shape[0] == image.shape[0]) or \
               not (vp_pos.shape[1] == vp_dir.shape[1] == image.shape[1]):
                msg = "Bad lens shape / direction for %s\n" % (self.lens_type)
                msg += "Shapes: (%s - %s - %s) and (%s - %s - %s)" % (
                    vp_pos.shape[0], vp_dir.shape[0], image.shape[0],
                    vp_pos.shape[1], vp_dir.shape[1], image.shape[1])
                raise RuntimeError(msg)

            if camera_data is not None and self.lens_type == 'perspective':
                self.extent_function = lenses.calculate_extent_perspective
            else:
                self.extent_function = lenses.calculate_extent_null
            self.\
                vector_function = lenses.generate_vector_info_null

        # These assignments are so we can track the objects and prevent their
        # de-allocation from reference counts.  Note that we do this to the
        # "atleast_3d" versions.  Also, note that we re-assign the input
        # arguments.
        self.vp_pos = vp_pos
        self.vp_dir = vp_dir
        self.image = self.aimage = image
        self.acenter = center
        self.center = <np.float64_t *> center.data
        self.ax_vec = x_vec
        self.x_vec = <np.float64_t *> x_vec.data
        self.ay_vec = y_vec
        self.y_vec = <np.float64_t *> y_vec.data
        self.zbuffer = zbuffer
        self.azbuffer = np.asarray(zbuffer)
        self.image_used = image_used
        self.aimage_used = np.asarray(image_used)
        self.mesh_lines = mesh_lines
        self.amesh_lines = np.asarray(mesh_lines)
        self.nv[0] = image.shape[0]
        self.nv[1] = image.shape[1]
        for i in range(4): self.bounds[i] = bounds[i]
        self.pdx = (bounds[1] - bounds[0])/self.nv[0]
        self.pdy = (bounds[3] - bounds[2])/self.nv[1]
        for i in range(3):
            self.width[i] = width[i]

    def __call__(self, PartitionedGrid pg, **kwa):
        if self.volume_method == 'KDTree':
            return self.cast_through_kdtree(pg, **kwa)
        elif self.volume_method == 'Octree':
            return self.cast_through_octree(pg, **kwa)
        else:
            raise NotImplementedError(
                'Volume rendering has not been implemented for method: "%s"' %
                self.volume_method
            )

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def cast_through_kdtree(self, PartitionedGrid pg, int num_threads = 0):
        # This routine will iterate over all of the vectors and cast each in
        # turn.  Might benefit from a more sophisticated intersection check,
        # like http://courses.csusm.edu/cs697exz/ray_box.htm
        cdef int vi, vj, hit, i, j
        cdef VolumeContainer *vc = pg.container
        self.setup(pg)
        cdef np.float64_t *v_pos
        cdef np.float64_t *v_dir
        cdef np.float64_t max_t
        hit = 0
        cdef np.int64_t nx, ny, size
        cdef np.int64_t iter[4]
        self.extent_function(self, vc, iter)
        iter[0] = i64clip(iter[0]-1, 0, self.nv[0])
        iter[1] = i64clip(iter[1]+1, 0, self.nv[0])
        iter[2] = i64clip(iter[2]-1, 0, self.nv[1])
        iter[3] = i64clip(iter[3]+1, 0, self.nv[1])
        nx = (iter[1] - iter[0])
        ny = (iter[3] - iter[2])
        size = nx * ny
        cdef ImageAccumulator *idata
        cdef np.float64_t width[3]
        cdef int chunksize = 100
        for i in range(3):
            width[i] = self.width[i]
        with nogil, parallel(num_threads = num_threads):
            idata = <ImageAccumulator *> malloc(sizeof(ImageAccumulator))
            idata.supp_data = self.supp_data
            v_pos = <np.float64_t *> malloc(3 * sizeof(np.float64_t))
            v_dir = <np.float64_t *> malloc(3 * sizeof(np.float64_t))
            for j in prange(size, schedule="static", chunksize=chunksize):
                vj = j % ny
                vi = (j - vj) / ny + iter[0]
                vj = vj + iter[2]
                # Dynamically calculate the position
                self.vector_function(self, vi, vj, width, v_dir, v_pos)
                for i in range(Nch):
                    idata.rgba[i] = self.image[vi, vj, i]
                max_t = fclip(self.zbuffer[vi, vj], 0.0, 1.0)
                walk_volume(vc, v_pos, v_dir, self.sample,
                            (<void *> idata), NULL, max_t)
                if (j % (10*chunksize)) == 0:
                    with gil:
                        PyErr_CheckSignals()
                for i in range(Nch):
                    self.image[vi, vj, i] = idata.rgba[i]
            idata.supp_data = NULL
            free(idata)
            free(v_pos)
            free(v_dir)
        return hit

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def cast_through_octree(self, PartitionedGrid pg, _OctreeRayTracing oct, int num_threads = 0):
        cdef RayInfo[int]* ri
        self.setup(pg)

        cdef sampler_function* sampler = <sampler_function*> self.sample

        cdef np.int64_t nx, ny, size
        cdef VolumeContainer *vc
        cdef ImageAccumulator *idata

        nx = self.nv[0]
        ny = self.nv[1]
        size = nx * ny

        cdef int i, j, k, vi, vj, icell
        cdef int[3] index = [0, 0, 0]
        cdef int chunksize = 100

        cdef int n_fields = pg.container.n_fields

        cdef np.float64_t vp_dir_len

        mylog.debug('Integrating rays')
        with nogil, parallel(num_threads=num_threads):
            idata = <ImageAccumulator *> malloc(sizeof(ImageAccumulator))
            idata.supp_data = self.supp_data

            ri = new RayInfo[int]()

            vc = <VolumeContainer*> malloc(sizeof(VolumeContainer))
            vc.n_fields = 1
            vc.data = <np.float64_t**> malloc(sizeof(np.float64_t*))
            vc.mask = <np.uint8_t*> malloc(8*sizeof(np.uint8_t))
            # The actual dimensions are 2x2x2, but the sampler
            # assumes vertex-centred data for a 1x1x1 lattice (i.e.
            # 2^3 vertices)
            vc.dims[0] = 1
            vc.dims[1] = 1
            vc.dims[2] = 1
            for j in prange(size, schedule='static', chunksize=chunksize):
                vj = j % ny
                vi = (j - vj) / ny

                vp_dir_len = sqrt(
                    self.vp_dir[vi, vj, 0]**2 +
                    self.vp_dir[vi, vj, 1]**2 +
                    self.vp_dir[vi, vj, 2]**2)

                # Cast ray
                oct.oct.cast_ray(&self.vp_pos[vi, vj, 0], &self.vp_dir[vi, vj, 0],
                                 ri.keys, ri.t)
                # Contains the ordered indices of the cells hit by the ray
                # and the entry/exit t values
                if ri.keys.size() == 0:
                    continue

                for i in range(Nch):
                    idata.rgba[i] = 0
                for i in range(8):
                    vc.mask[i] = 1

                # Iterate over cells
                for i in range(ri.keys.size()):
                    icell = ri.keys[i]
                    for k in range(n_fields):
                        vc.data[k] = &pg.container.data[k][14*icell]

                    # Fill the volume container with the current boundaries
                    for k in range(3):
                        vc.left_edge[k] = pg.container.data[0][14*icell+8+k]
                        vc.right_edge[k] = pg.container.data[0][14*icell+11+k]
                        vc.dds[k] = (vc.right_edge[k] - vc.left_edge[k])
                        vc.idds[k] = 1/vc.dds[k]
                    # Now call the sampler
                    sampler(
                        vc,
                        &self.vp_pos[vi, vj, 0],
                        &self.vp_dir[vi, vj, 0],
                        ri.t[2*i  ]/vp_dir_len,
                        ri.t[2*i+1]/vp_dir_len,
                        index,
                        <void *> idata
                        )
                for i in range(Nch):
                    self.image[vi, vj, i] = idata.rgba[i]

                # Empty keys and t
                ri.keys.clear()
                ri.t.clear()

            del ri
            free(vc.data)
            free(vc.mask)
            free(vc)
            idata.supp_data = NULL
            free(idata)

        mylog.debug('Done integration')


    cdef void setup(self, PartitionedGrid pg):
        return

    @staticmethod
    cdef void sample(
                 VolumeContainer *vc,
                 np.float64_t v_pos[3],
                 np.float64_t v_dir[3],
                 np.float64_t enter_t,
                 np.float64_t exit_t,
                 int index[3],
                 void *data) nogil:
        return

    def ensure_code_unit_params(self, params):
        for param_name in ['center', 'vp_pos', 'vp_dir', 'width']:
            param = params[param_name]
            if hasattr(param, 'in_units'):
                params[param_name] = param.in_units('code_length')
        bounds = params['bounds']
        if hasattr(bounds[0], 'units'):
            params['bounds'] = tuple(b.in_units('code_length').d for b in bounds)

        return params

cdef class ProjectionSampler(ImageSampler):

    @staticmethod
    cdef void sample(
                 VolumeContainer *vc,
                 np.float64_t v_pos[3],
                 np.float64_t v_dir[3],
                 np.float64_t enter_t,
                 np.float64_t exit_t,
                 int index[3],
                 void *data) nogil:
        cdef ImageAccumulator *im = <ImageAccumulator *> data
        cdef int i
        cdef np.float64_t dl = (exit_t - enter_t)
        cdef int di = (index[0]*vc.dims[1]+index[1])*vc.dims[2]+index[2]
        for i in range(imin(4, vc.n_fields)):
            im.rgba[i] += vc.data[i][di] * dl


cdef class InterpolatedProjectionSampler(ImageSampler):
    def __cinit__(self,
                  np.ndarray vp_pos,
                  np.ndarray vp_dir,
                  np.ndarray[np.float64_t, ndim=1] center,
                  bounds,
                  np.ndarray[np.float64_t, ndim=3] image,
                  np.ndarray[np.float64_t, ndim=1] x_vec,
                  np.ndarray[np.float64_t, ndim=1] y_vec,
                  np.ndarray[np.float64_t, ndim=1] width,
                  str volume_method,
                  n_samples = 10,
                  **kwargs
        ):
        ImageSampler.__init__(self, vp_pos, vp_dir, center, bounds, image,
                               x_vec, y_vec, width, volume_method, **kwargs)
        # Now we handle tf_obj
        self.vra = <VolumeRenderAccumulator *> \
            malloc(sizeof(VolumeRenderAccumulator))
        self.vra.n_samples = n_samples
        self.supp_data = <void *> self.vra

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @staticmethod
    cdef void sample(
                     VolumeContainer *vc,
                     np.float64_t v_pos[3],
                     np.float64_t v_dir[3],
                     np.float64_t enter_t,
                     np.float64_t exit_t,
                     int index[3],
                     void *data) nogil:
        cdef ImageAccumulator *im = <ImageAccumulator *> data
        cdef VolumeRenderAccumulator *vri = <VolumeRenderAccumulator *> \
                im.supp_data
        # we assume this has vertex-centered data.
        cdef int offset = index[0] * (vc.dims[1] + 1) * (vc.dims[2] + 1) \
                        + index[1] * (vc.dims[2] + 1) + index[2]
        cdef np.float64_t dp[3]
        cdef np.float64_t ds[3]
        cdef np.float64_t dt = (exit_t - enter_t) / vri.n_samples
        cdef np.float64_t dvs[6]
        for i in range(3):
            dp[i] = (enter_t + 0.5 * dt) * v_dir[i] + v_pos[i]
            dp[i] -= index[i] * vc.dds[i] + vc.left_edge[i]
            dp[i] *= vc.idds[i]
            ds[i] = v_dir[i] * vc.idds[i] * dt
        for i in range(vri.n_samples):
            for j in range(vc.n_fields):
                dvs[j] = offset_interpolate(vc.dims, dp,
                        vc.data[j] + offset)
            for j in range(imin(3, vc.n_fields)):
                im.rgba[j] += dvs[j] * dt
            for j in range(3):
                dp[j] += ds[j]


cdef class VolumeRenderSampler(ImageSampler):
    def __cinit__(self,
                  np.ndarray vp_pos,
                  np.ndarray vp_dir,
                  np.ndarray[np.float64_t, ndim=1] center,
                  bounds,
                  np.ndarray[np.float64_t, ndim=3] image,
                  np.ndarray[np.float64_t, ndim=1] x_vec,
                  np.ndarray[np.float64_t, ndim=1] y_vec,
                  np.ndarray[np.float64_t, ndim=1] width,
                  str volume_method,
                  tf_obj,
                  n_samples = 10,
                  **kwargs
        ):
        ImageSampler.__init__(self, vp_pos, vp_dir, center, bounds, image,
                               x_vec, y_vec, width, volume_method, **kwargs)
        cdef int i
        cdef np.ndarray[np.float64_t, ndim=1] temp
        # Now we handle tf_obj
        self.vra = <VolumeRenderAccumulator *> \
            malloc(sizeof(VolumeRenderAccumulator))
        self.vra.fits = <FieldInterpolationTable *> \
            malloc(sizeof(FieldInterpolationTable) * 6)
        self.vra.n_fits = tf_obj.n_field_tables
        assert(self.vra.n_fits <= 6)
        self.vra.grey_opacity = getattr(tf_obj, "grey_opacity", 0)
        self.vra.n_samples = n_samples
        self.my_field_tables = []
        for i in range(self.vra.n_fits):
            temp = tf_obj.tables[i].y
            FIT_initialize_table(&self.vra.fits[i],
                      temp.shape[0],
                      <np.float64_t *> temp.data,
                      tf_obj.tables[i].x_bounds[0],
                      tf_obj.tables[i].x_bounds[1],
                      tf_obj.field_ids[i], tf_obj.weight_field_ids[i],
                      tf_obj.weight_table_ids[i])
            self.my_field_tables.append((tf_obj.tables[i],
                                         tf_obj.tables[i].y))
        for i in range(6):
            self.vra.field_table_ids[i] = tf_obj.field_table_ids[i]
        self.supp_data = <void *> self.vra

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @staticmethod
    cdef void sample(
                     VolumeContainer *vc,
                     np.float64_t v_pos[3],
                     np.float64_t v_dir[3],
                     np.float64_t enter_t,
                     np.float64_t exit_t,
                     int index[3],
                     void *data) nogil:
        cdef ImageAccumulator *im = <ImageAccumulator *> data
        cdef VolumeRenderAccumulator *vri = <VolumeRenderAccumulator *> \
                im.supp_data
        # we assume this has vertex-centered data.
        cdef int offset = index[0] * (vc.dims[1] + 1) * (vc.dims[2] + 1) \
                        + index[1] * (vc.dims[2] + 1) + index[2]
        cdef int cell_offset = index[0] * (vc.dims[1]) * (vc.dims[2]) \
                        + index[1] * (vc.dims[2]) + index[2]
        if vc.mask[cell_offset] != 1:
            return
        cdef np.float64_t dp[3]
        cdef np.float64_t ds[3]
        cdef np.float64_t dt = (exit_t - enter_t) / vri.n_samples
        cdef np.float64_t dvs[6]
        for i in range(3):
            dp[i] = (enter_t + 0.5 * dt) * v_dir[i] + v_pos[i]
            dp[i] -= index[i] * vc.dds[i] + vc.left_edge[i]
            dp[i] *= vc.idds[i]
            ds[i] = v_dir[i] * vc.idds[i] * dt
        for i in range(vri.n_samples):
            for j in range(vc.n_fields):
                dvs[j] = offset_interpolate(vc.dims, dp,
                        vc.data[j] + offset)
            FIT_eval_transfer(dt, dvs, im.rgba, vri.n_fits,
                    vri.fits, vri.field_table_ids, vri.grey_opacity)
            for j in range(3):
                dp[j] += ds[j]

    def __dealloc__(self):
        for i in range(self.vra.n_fits):
            free(self.vra.fits[i].d0)
            free(self.vra.fits[i].dy)
        free(self.vra.fits)
        free(self.vra)

cdef class LightSourceRenderSampler(ImageSampler):
    def __cinit__(self,
                  np.ndarray vp_pos,
                  np.ndarray vp_dir,
                  np.ndarray[np.float64_t, ndim=1] center,
                  bounds,
                  np.ndarray[np.float64_t, ndim=3] image,
                  np.ndarray[np.float64_t, ndim=1] x_vec,
                  np.ndarray[np.float64_t, ndim=1] y_vec,
                  np.ndarray[np.float64_t, ndim=1] width,
                  str volume_method,
                  tf_obj,
                  n_samples = 10,
                  light_dir=[1.,1.,1.],
                  light_rgba=[1.,1.,1.,1.],
                  **kwargs):
        ImageSampler.__init__(self, vp_pos, vp_dir, center, bounds, image,
                               x_vec, y_vec, width, volume_method, **kwargs)
        cdef int i
        cdef np.ndarray[np.float64_t, ndim=1] temp
        # Now we handle tf_obj
        self.vra = <VolumeRenderAccumulator *> \
            malloc(sizeof(VolumeRenderAccumulator))
        self.vra.fits = <FieldInterpolationTable *> \
            malloc(sizeof(FieldInterpolationTable) * 6)
        self.vra.n_fits = tf_obj.n_field_tables
        assert(self.vra.n_fits <= 6)
        self.vra.grey_opacity = getattr(tf_obj, "grey_opacity", 0)
        self.vra.n_samples = n_samples
        self.vra.light_dir = <np.float64_t *> malloc(sizeof(np.float64_t) * 3)
        self.vra.light_rgba = <np.float64_t *> malloc(sizeof(np.float64_t) * 4)
        light_dir /= np.sqrt(light_dir[0] * light_dir[0] +
                             light_dir[1] * light_dir[1] +
                             light_dir[2] * light_dir[2])
        for i in range(3):
            self.vra.light_dir[i] = light_dir[i]
        for i in range(4):
            self.vra.light_rgba[i] = light_rgba[i]
        self.my_field_tables = []
        for i in range(self.vra.n_fits):
            temp = tf_obj.tables[i].y
            FIT_initialize_table(&self.vra.fits[i],
                      temp.shape[0],
                      <np.float64_t *> temp.data,
                      tf_obj.tables[i].x_bounds[0],
                      tf_obj.tables[i].x_bounds[1],
                      tf_obj.field_ids[i], tf_obj.weight_field_ids[i],
                      tf_obj.weight_table_ids[i])
            self.my_field_tables.append((tf_obj.tables[i],
                                         tf_obj.tables[i].y))
        for i in range(6):
            self.vra.field_table_ids[i] = tf_obj.field_table_ids[i]
        self.supp_data = <void *> self.vra

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @staticmethod
    cdef void sample(
                     VolumeContainer *vc,
                     np.float64_t v_pos[3],
                     np.float64_t v_dir[3],
                     np.float64_t enter_t,
                     np.float64_t exit_t,
                     int index[3],
                     void *data) nogil:
        cdef ImageAccumulator *im = <ImageAccumulator *> data
        cdef VolumeRenderAccumulator *vri = <VolumeRenderAccumulator *> \
                im.supp_data
        # we assume this has vertex-centered data.
        cdef int offset = index[0] * (vc.dims[1] + 1) * (vc.dims[2] + 1) \
                        + index[1] * (vc.dims[2] + 1) + index[2]
        cdef np.float64_t dp[3]
        cdef np.float64_t ds[3]
        cdef np.float64_t dt = (exit_t - enter_t) / vri.n_samples
        cdef np.float64_t dvs[6]
        cdef np.float64_t *grad
        grad = <np.float64_t *> malloc(3 * sizeof(np.float64_t))
        for i in range(3):
            dp[i] = (enter_t + 0.5 * dt) * v_dir[i] + v_pos[i]
            dp[i] -= index[i] * vc.dds[i] + vc.left_edge[i]
            dp[i] *= vc.idds[i]
            ds[i] = v_dir[i] * vc.idds[i] * dt
        for i in range(vri.n_samples):
            for j in range(vc.n_fields):
                dvs[j] = offset_interpolate(vc.dims, dp,
                        vc.data[j] + offset)
            eval_gradient(vc.dims, dp, vc.data[0] + offset, grad)
            FIT_eval_transfer_with_light(dt, dvs, grad,
                    vri.light_dir, vri.light_rgba,
                    im.rgba, vri.n_fits,
                    vri.fits, vri.field_table_ids, vri.grey_opacity)
            for j in range(3):
                dp[j] += ds[j]
        free(grad)


    def __dealloc__(self):
        for i in range(self.vra.n_fits):
            free(self.vra.fits[i].d0)
            free(self.vra.fits[i].dy)
        free(self.vra.light_dir)
        free(self.vra.light_rgba)
        free(self.vra.fits)
        free(self.vra)



import tempfile

from yt.testing import fake_particle_ds, requires_module


@requires_module("firefly_api")
def test_firefly_JSON_object():
    tmpdir = tempfile.mkdtemp()

    ds = fake_particle_ds()
    ad = ds.all_data()
    reader = ad.create_firefly_object(
        path_to_firefly=tmpdir,
        velocity_units="cm/s",
        coordinate_units="cm",
        dataset_name="test",
    )
    reader.dumpToJSON()



# distutils: libraries = STD_LIBS
# Note that we used to include the empty c_kdtree.cpp file, but that seems to break cythonize.
# distutils: sources = yt/utilities/lib/cykdtree/c_utils.cpp
# distutils: depends = yt/utilities/lib/cykdtree/c_kdtree.hpp, yt/utilities/lib/cykdtree/c_utils.hpp
# distutils: language = c++
# distutils: extra_compile_args = CPP03_FLAG
import cython
import numpy as np

cimport numpy as np
from cpython cimport bool as pybool
from cython.operator cimport dereference
from libc.stdint cimport int32_t, int64_t, uint32_t, uint64_t
from libc.stdlib cimport free, malloc
from libcpp cimport bool as cbool


cdef class PyNode:
    r"""A container for leaf info.

    Attributes:
        npts (np.uint64_t): Number of points in this node.
        ndim (np.uint32_t): Number of dimensions in domain.
        num_leaves (np.uint32_t): Number of leaves in the tree containing this
            node.
        start_idx (np.uint64_t): Index where indices for this node begin.
        stop_idx (np.uint64_t): One passed the end of indices for this node.
        left_edge (np.ndarray of float64): Minimum bounds of this node in each
            dimension.
        right_edge (np.ndarray of float64): Maximum bounds of this node in each
            dimension.
        periodic_left (np.ndarray of bool): Periodicity of minimum bounds.
        periodic_right (np.ndarray of bool): Periodicity of maximum bounds.
        domain_width (np.ndarray of float64): Width of the total domain in each
            dimension.
        left_neighbors (list of lists): Indices of neighbor leaves at the
            minimum bounds in each dimension.
        right_neighbors (list of lists): Indices of neighbor leaves at the
            maximum bounds in each dimension.

    """

    cdef void _init_node(self, Node* node, uint32_t num_leaves,
                         double *domain_width):
        cdef np.uint32_t i, j
        self._node = node
        self.id = node.leafid
        self.npts = node.children
        self.ndim = node.ndim
        self.num_leaves = num_leaves
        self.start_idx = node.left_idx
        self.stop_idx = (node.left_idx + node.children)
        self._domain_width = domain_width
        self.left_neighbors = [None for i in range(self.ndim)]
        self.right_neighbors = [None for i in range(self.ndim)]
        for i in range(self.ndim):
            self.left_neighbors[i] = [node.left_neighbors[i][j] for j in
                                      range(node.left_neighbors[i].size())]
            self.right_neighbors[i] = [node.right_neighbors[i][j] for j in
                                       range(node.right_neighbors[i].size())]

    def __cinit__(self):
        # Initialize everthing to NULL/0/None to prevent seg fault
        self._node = NULL
        self.id = 0
        self.npts = 0
        self.ndim = 0
        self.num_leaves = 0
        self.start_idx = 0
        self.stop_idx = 0
        self._domain_width = NULL
        self.left_neighbors = None
        self.right_neighbors = None

    def __init__(self):
        pass

    def __repr__(self):
        nchars = 1 + len(str(self.__class__.__name__))
        return ('%s(id=%i, npts=%i, start_idx=%i, stop_idx=%i,\n' +
                ' ' * nchars + 'left_edge=%s,\n' +
                ' ' * nchars + 'right_edge=%s)') % (
            self.__class__.__name__,
            self.id,
            self.npts,
            self.start_idx,
            self.stop_idx,
            self.left_edge,
            self.right_edge,
        )

    @property
    def periodic_left(self):
        cdef cbool[:] view = <cbool[:self.ndim]> self._node.periodic_left
        return np.asarray(view)
    @property
    def periodic_right(self):
        cdef cbool[:] view = <cbool[:self.ndim]> self._node.periodic_right
        return np.asarray(view)
    @property
    def left_edge(self):
        cdef np.float64_t[:] view = <np.float64_t[:self.ndim]> self._node.left_edge
        return np.asarray(view)
    @property
    def right_edge(self):
        cdef np.float64_t[:] view = <np.float64_t[:self.ndim]> self._node.right_edge
        return np.asarray(view)
    @property
    def domain_width(self):
        cdef np.float64_t[:] view = <np.float64_t[:self.ndim]> self._domain_width
        return np.asarray(view)

    @property
    def slice(self):
        """slice: Slice of kdtree indices contained by this node."""
        return slice(self.start_idx, self.stop_idx)

    @property
    def neighbors(self):
        """list of int: Indices of all neighboring leaves including this
        leaf."""
        cdef np.uint32_t i
        cdef object out
        cdef vector[uint32_t] vout = self._node.all_neighbors
        out = [vout[i] for i in range(<np.uint32_t>vout.size())]
        return out

    def assert_equal(self, PyNode solf):
        """Assert that node properties are equal."""
        np.testing.assert_equal(self.npts, solf.npts)
        np.testing.assert_equal(self.ndim, solf.ndim)
        np.testing.assert_equal(self.num_leaves, solf.num_leaves)
        np.testing.assert_equal(self.id, solf.id)
        np.testing.assert_equal(self.start_idx, solf.start_idx)
        np.testing.assert_equal(self.stop_idx, solf.stop_idx)
        np.testing.assert_array_equal(self.left_edge, solf.left_edge)
        np.testing.assert_array_equal(self.right_edge, solf.right_edge)
        np.testing.assert_array_equal(self.periodic_left, solf.periodic_left)
        np.testing.assert_array_equal(self.periodic_right, solf.periodic_right)
        for i in range(self.ndim):
            np.testing.assert_equal(self.left_neighbors[i], solf.left_neighbors[i])
            np.testing.assert_equal(self.right_neighbors[i], solf.right_neighbors[i])
        np.testing.assert_equal(self.neighbors, solf.neighbors)


cdef class PyKDTree:
    r"""Construct a KDTree for a set of points.

    Args:
        pts (np.ndarray of double): (n,m) array of n coordinates in a
            m-dimensional domain.
        left_edge (np.ndarray of double): (m,) domain minimum in each dimension.
        right_edge (np.ndarray of double): (m,) domain maximum in each dimension.
        periodic (bool or np.ndarray of bool, optional): Truth of the domain
            periodicity overall (if bool), or in each dimension (if np.ndarray).
            Defaults to `False`.
        leafsize (int, optional): The maximum number of points that should be in
            a leaf. Defaults to 10000.
        nleaves (int, optional): The number of leaves that should be in the
            resulting tree. If greater than 0, leafsize is adjusted to produce a
            tree with 2**(ceil(log2(nleaves))) leaves. The leafsize keyword
            argument is ignored if nleaves is greater zero. Defaults to 0.
        data_version (int, optional): An optional user-provided integer that
            can be used to uniquely identify the data used to generate the
            KDTree. This is useful if you save the kdtree to disk and restore
            it later and need to verify that the underlying data is the same.
        use_sliding_midpoint (bool, optional): If True, the sliding midpoint
            rule is used to perform splits. Otherwise, the median is used.
            Defaults to False.

    Raises:
        ValueError: If `leafsize < 2`. This currectly segfaults.

    Attributes:
        npts (uint64): Number of points in the tree.
        ndim (uint32): Number of dimensions points occupy.
        data_version (int64): User-provided version number (defaults to 0)
        num_leaves (uint32): Number of leaves in the tree.
        leafsize (uint32): Maximum number of points a leaf can have.
        leaves (list of `cykdtree.PyNode`): Tree leaves.
        idx (np.ndarray of uint64): Indices sorting the points by leaf.
        left_edge (np.ndarray of double): (m,) domain minimum in each dimension.
        right_edge (np.ndarray of double): (m,) domain maximum in each dimension.
        domain_width (np.ndarray of double): (m,) domain width in each dimension.
        periodic (np.ndarray of bool): Truth of domain periodicity in each
            dimension.

    """

    cdef void _init_tree(self, KDTree* tree):
        self._tree = tree
        self.ndim = tree.ndim
        self.data_version = tree.data_version
        self.npts = tree.npts
        self.num_leaves = tree.num_leaves
        self.leafsize = tree.leafsize
        self._make_leaves()
        self._idx = np.empty(self.npts, 'uint64')
        cdef uint64_t i
        for i in range(self.npts):
            self._idx[i] = tree.all_idx[i]

    def __cinit__(self):
        # Initialize everthing to NULL/0/None to prevent seg fault
        self._tree = NULL
        self.npts = 0
        self.ndim = 0
        self.num_leaves = 0
        self.leafsize = 0
        self._left_edge = NULL
        self._right_edge = NULL
        self._periodic = NULL
        self.leaves = None
        self._idx = None

    def __init__(self, np.ndarray[double, ndim=2] pts = None,
                 left_edge = None,
                 right_edge = None,
                 periodic = False,
                 int leafsize = 10000,
                 int nleaves = 0,
                 data_version = None,
                 use_sliding_midpoint = False):
        # Return with nothing set if points not provided
        if pts is None:
            return
        # Set leafsize of number of leaves provided
        if nleaves > 0:
            nleaves = <int>(2**np.ceil(np.log2(<float>nleaves)))
            leafsize = pts.shape[0]/nleaves + 1
        if (leafsize < 2):
            # This is here to prevent segfault. The cpp code needs modified to
            # support leafsize = 1
            raise ValueError("'leafsize' cannot be smaller than 2.")
        if left_edge is None:
            left_edge = np.min(pts, axis=0)
        else:
            left_edge = np.array(left_edge)
        if right_edge is None:
            right_edge = np.max(pts, axis=0)
        else:
            right_edge = np.array(right_edge)
        if data_version is None:
            data_version = 0
        self.data_version = data_version
        cdef uint32_t k,i,j
        self.npts = <uint64_t>pts.shape[0]
        self.ndim = <uint32_t>pts.shape[1]
        assert(left_edge.size == self.ndim)
        assert(right_edge.size == self.ndim)
        self.leafsize = leafsize
        self._left_edge = <double *>malloc(self.ndim*sizeof(double))
        self._right_edge = <double *>malloc(self.ndim*sizeof(double))
        self._periodic = <cbool *>malloc(self.ndim*sizeof(cbool));
        for i in range(self.ndim):
            self._left_edge[i] = left_edge[i]
            self._right_edge[i] = right_edge[i]
        if isinstance(periodic, pybool):
            for i in range(self.ndim):
                self._periodic[i] = <cbool>periodic
        else:
            for i in range(self.ndim):
                self._periodic[i] = <cbool>periodic[i]
        # Create tree and leaves
        self._make_tree(&pts[0,0], <cbool>use_sliding_midpoint)
        self._make_leaves()

    def __dealloc__(self):
        if self._tree != NULL:
            del self._tree
        if self._left_edge != NULL:
            free(self._left_edge)
        if self._right_edge != NULL:
            free(self._right_edge)
        if self._periodic != NULL:
            free(self._periodic)

    def assert_equal(self, PyKDTree solf, pybool strict_idx = True):
        r"""Compare this tree to another tree.

        Args:
            solf (PyKDTree): Another KDTree to compare with this one.
            strict_idx (bool, optional): If True, the index vectors are
                compared for equality element by element. If False,
                corresponding leaves must contain the same indices, but they
                can be in any order. Defaults to True.

        Raises:
            AssertionError: If there are missmatches between any of the two
                trees' parameters.

        """
        np.testing.assert_equal(self.npts, solf.npts)
        np.testing.assert_equal(self.ndim, solf.ndim)
        np.testing.assert_equal(self.data_version, solf.data_version)
        np.testing.assert_equal(self.leafsize, solf.leafsize)
        np.testing.assert_equal(self.num_leaves, solf.num_leaves)
        np.testing.assert_array_equal(self.left_edge, solf.left_edge)
        np.testing.assert_array_equal(self.right_edge, solf.right_edge)
        np.testing.assert_array_equal(self.periodic, solf.periodic)
        # Compare index at the leaf level since we only care that the leaves
        # contain the same points
        if strict_idx:
            np.testing.assert_array_equal(self._idx, solf._idx)
        for i in range(self.num_leaves):
            self.leaves[i].assert_equal(solf.leaves[i])
            if not strict_idx:
                np.testing.assert_array_equal(
                    np.sort(self._idx[self.leaves[i].slice]),
                    np.sort(solf._idx[solf.leaves[i].slice]))

    cdef void _make_tree(self, double *pts, bool use_sliding_midpoint):
        r"""Carry out creation of KDTree at C++ level."""
        cdef uint64_t[:] idx = np.arange(self.npts).astype('uint64')
        self._tree = new KDTree(pts, &idx[0], self.npts, self.ndim, self.leafsize,
                                self._left_edge, self._right_edge, self._periodic,
                                self.data_version, use_sliding_midpoint)
        self._idx = idx

    cdef void _make_leaves(self):
        r"""Create a list of Python leaf objects from C++ leaves."""
        self.num_leaves = <uint32_t>self._tree.leaves.size()
        self.leaves = [None for _ in xrange(self.num_leaves)]
        cdef Node* leafnode
        cdef PyNode leafnode_py
        cdef object leaf_neighbors = None
        for k in xrange(self.num_leaves):
            leafnode = self._tree.leaves[k]
            leafnode_py = PyNode()
            leafnode_py._init_node(leafnode, self.num_leaves,
                                   self._tree.domain_width)
            self.leaves[leafnode.leafid] = leafnode_py

    @property
    def left_edge(self):
        cdef np.float64_t[:] view = <np.float64_t[:self.ndim]> self._tree.domain_left_edge
        return np.asarray(view)
    @property
    def right_edge(self):
        cdef np.float64_t[:] view = <np.float64_t[:self.ndim]> self._tree.domain_right_edge
        return np.asarray(view)
    @property
    def domain_width(self):
        cdef np.float64_t[:] view = <np.float64_t[:self.ndim]> self._tree.domain_width
        return np.asarray(view)
    @property
    def periodic(self):
        cdef cbool[:] view = <cbool[:self.ndim]> self._tree.periodic
        # return np.asarray(view)
        cdef object out = np.empty(self.ndim, 'bool')
        cdef np.uint32_t i
        for i in range(self.ndim):
            out[i] = view[i]
        return out

    def leaf_idx(self, np.uint32_t leafid):
        r"""Get array of indices for points in a leaf.

        Args:
            leafid (np.uint32_t): Unique index of the leaf in question.

        Returns:
            np.ndarray of np.uint64_t: Indices of points belonging to leaf.

        """
        cdef np.ndarray[np.uint64_t] out = self._idx[self.leaves[leafid].slice]
        return out

    cdef np.ndarray[np.uint32_t, ndim=1] _get_neighbor_ids(self, np.ndarray[double, ndim=1] pos):
        cdef np.uint32_t i
        cdef vector[uint32_t] vout = self._tree.get_neighbor_ids(&pos[0]);
        cdef np.ndarray[np.uint32_t, ndim=1] out = np.empty(vout.size(), 'uint32')
        for i in xrange(vout.size()):
            out[i] = vout[i]
        return out

    @property
    def idx(self):
        return np.asarray(self._idx)

    def get_neighbor_ids(self, np.ndarray[double, ndim=1] pos):
        r"""Return the IDs of leaves containing & neighboring a given position.

        Args:
            pos (np.ndarray of double): Coordinates.

        Returns:
            np.ndarray of uint32: Leaves containing/neighboring `pos`.

        Raises:
            ValueError: If pos is not contained withing the KDTree.

        """
        return self._get_neighbor_ids(pos)

    cdef np.ndarray[np.uint32_t, ndim=1] _get_neighbor_ids_3(self, np.float64_t pos[3]):
        cdef np.uint32_t i
        cdef vector[uint32_t] vout = self._tree.get_neighbor_ids(&pos[0]);
        cdef np.ndarray[np.uint32_t, ndim=1] out = np.empty(vout.size(), 'uint32')
        for i in xrange(vout.size()):
            out[i] = vout[i]
        return out

    cdef PyNode _get(self, np.ndarray[double, ndim=1] pos):
        assert(<uint32_t>len(pos) == self.ndim)
        cdef Node* leafnode = self._tree.search(&pos[0])
        if leafnode == NULL:
            raise ValueError("Position is not within the kdtree root node.")
        cdef PyNode out = self.leaves[leafnode.leafid]
        return out

    def get(self, np.ndarray[double, ndim=1] pos):
        r"""Return the leaf containing a given position.

        Args:
            pos (np.ndarray of double): Coordinates.

        Returns:
            :class:`cykdtree.PyNode`: Leaf containing `pos`.

        Raises:
            ValueError: If pos is not contained withing the KDTree.

        """
        return self._get(pos)

    def consolidate_edges(self):
        r"""Return arrays of the left and right edges for all leaves in the
        tree on each process.

        Returns:
            tuple(np.ndarray of double, np.ndarray of double): The left (first
                array) and right (second array) edges of each leaf (1st array
                dimension), in each dimension (2nd array dimension).

        """
        cdef np.ndarray[np.float64_t, ndim=2] leaves_le
        cdef np.ndarray[np.float64_t, ndim=2] leaves_re
        leaves_le = np.empty((self.num_leaves, self.ndim), 'float64')
        leaves_re = np.empty((self.num_leaves, self.ndim), 'float64')
        self._tree.consolidate_edges(&leaves_le[0,0], &leaves_re[0,0])
        return (leaves_le, leaves_re)

    def save(self, str filename):
        r"""Saves the PyKDTree to disk as raw binary data.

        Note that this file may not necessarily be portable.

        Args:
            filename (string): Name of the file to serialize the kdtree to

        """
        cdef KDTree* my_tree = self._tree
        cdef ofstream* outputter = new ofstream(filename.encode('utf8'), binary)
        try:
            my_tree.serialize(dereference(outputter))
        finally:
            del outputter

    @classmethod
    def from_file(cls, str filename, data_version=None):
        r"""Create a PyKDTree from a binary file created by ``PyKDTree.save()``

        Note that loading a file created on another machine may create
        a corrupted PyKDTree instance.

        Args:
            filename (string): Name of the file to load the kdtree from
            data_version (int): A unique integer corresponding to the data
                                being loaded. If the loaded data_version does
                                not match the data_version supplied here then
                                an OSError is raised. Optional.

        Returns:
            :class:`cykdtree.PyKDTree`: A KDTree restored from the file

        """
        cdef ifstream* inputter = new ifstream(filename.encode(), binary)
        cdef PyKDTree ret = cls()
        if data_version is None:
            data_version = 0
        try:
            ret._init_tree(new KDTree(dereference(inputter)))
        finally:
            del inputter
        return ret
